/*************************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * Name: ORVA_GET_KRI_LOG._sas                                             
 *
 * Title: Construct the KRI_LOG macro that contains the string 
 *        X_1*B_1 + ... + X_n*B_n used to normalize the losses 
 *        according to their KRI values   
 *
 * Please see the accompanying doc for pdfs and cdfs of the distributions
 *
 * Parameters
 *
 * DATA      = Specifies the name of the input data set to be analyzed.
 *
 *
 **************************************************************************************/
%macro orva_get_kri_log(data=);

   data _gkl_data;
      set &data.;
      length NAME $ 100;
      array  KRIs Param_1_:;
      do i = 1 to dim(KRIs);
         NAME  = trim(substrn(vname(KRIs(i)),9));
         VALUE = put(KRIs(i), BEST20.);
         output;
               end;

      call symputx( "MY_INTERCEPT"  , Param_1  );
      
      keep
         NAME
         VALUE
         ;
   run;

   proc sql noprint;
      select
               NAME || "*" || VALUE,
               NAME,
               "SEFF_" || NAME,
               "AVG_SEFF_" || NAME || " 0",
               "Loss_" || NAME,
               NAME || " " || VALUE,
               "AVG_" || NAME
               

      into
               :KRI_LOG separated by " + " ,
               :KRI_LIST separated by " ",
               :KRI_EFFECT separated by " ",
               :KRI_EFFECT_RETAIN separated by " ",
               :KRI_LOSS_EFFECT separated by " ",
               :KRI_RETAIN separated by " ",
               :KRI_AVG_EFFECT separated by " "
      from
               _gkl_data
      ;
   quit;

   %orvu_delete_data( data = _gkl_data );

%mend  orva_get_kri_log;
