/*****************************************************************************************
 * Copyright (C), 2012
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *  NAME:    orva_init_global_macros.sas                                             
 *  PURPOSE: Second half of initial set up for the OpRiskVaR application.
 *           This macro is called by orva_init() shortly after connecting to the server.
 *           
 *  SUPPORT: jariou
 *  
 *  PARAMETERS: 
 *     infolib - where are all the informational datasets?
 *      
 *  NOTES: This macro uses informational distribution datasets created from the
 *         distributions.xml to create numerous global macro variables.
 *         
 *         It also loads (invokes) every distribution's init macro.
 *   
 *****************************************************************************************/
%macro orva_init_global_macros(infolib = work);
   %global ORV_ABORT;
   %global ORV_REASON;

   %global ORV_FALSE;
   %let    ORV_FALSE = 0;

   %global ORV_TRUE ;
   %let    ORV_TRUE  = 1;

   %global BASE_PROC_NLP_OPTIONS; 
   %let    BASE_PROC_NLP_OPTIONS = %str(COV=2 VARDEF=n pcov maxit=100 maxfu=10000 tech=tr gradcheck=none);

   %global LAST_PROC_NLP_OPTIONS; 
   %let    LAST_PROC_NLP_OPTIONS = %str(COV=2 VARDEF=n pcov maxit=25 maxfu=10000 tech=nrr gradcheck=none);

   %global PROC_NLP_OPTIONS; 
   %let    PROC_NLP_OPTIONS = %str(&BASE_PROC_NLP_OPTIONS.);

   /***************************************************************/
   /***             Setup DATASET_NAME Length                   ***/
   /***************************************************************/
   %global ORV_DATASET_NAME_LENGTH;
   %let ORV_DATASET_NAME_LENGTH = 41;

   proc sort data = &infolib..sparmnames;
      by distribution;
   run;

   proc transpose 
                data = &infolib..sparmnames 
                out  = _ifs_DistNames
                ;
      var DatasetName;
      by distribution;
   run;

   data _NULL_; 
      set _ifs_DistNames;
      array cols[*] col:;
      call symputX("MAX_SEV_PARAMS", dim(cols));
      stop;
   run;

   %do i = 1 %to &MAX_SEV_PARAMS;
      %global param_&i._init;
      %global param_&i.;   
   %end;

   %global Param_pnLimit;
   %global Param_xLimit; 
   
   /*- create global macro variable &dist.._PARAM_LIST_MINUS_1 -*/
   data _NULL_; 
      rc = filename("glob_mac",,"temp"); 
   run;

   data _NULL_;
      length tmpList $128;
      set &infolib..sparmnames; 
      by Distribution;
      file glob_mac;    

      if first.Distribution then do;
         tmpList = compress(Distribution || "_PARAM_LIST_MINUS_1");
         put '%global ' tmpList ";";
         Param_List_Minus_1 ="";
      end;
   run;

   %include glob_mac;

   /*- set global macro variable &dist.._PARAM_LIST_MINUS_1 -*/
   data _NULL_;
      set &infolib..sparmnames;
      by Distribution;
      length Param_List_Minus_1 $128;
      retain Param_List_Minus_1 "";

      if first.Distribution then do;
         Param_List_Minus_1 = "";
      end;
      else do;
         Param_List_Minus_1 = strip(Param_List_Minus_1)|| ", " || DatasetName;
      end;

      if last.Distribution then do;
         call symput(compress(Distribution || "_PARAM_LIST_MINUS_1"), Param_List_Minus_1);
      end;
   run;

   proc sort data = &infolib..sparmnames;
      by distribution;
   run;

   proc transpose 
                 data = &infolib..fparmnames 
                 out  = _ifs_DistNames
                 ;
      var DatasetName;
      by distribution;
   run;

   data _NULL_; 
      set _ifs_DistNames;
      array cols[*] col:;
      call symputX("MAX_FREQ_PARAMS", dim(cols));
      stop;
   run;
 
   %do i = 1 %to &MAX_FREQ_PARAMS;
      %global fparam_&i._init;
      %global fparam_&i.;   
   %end;

   %global SimDate;
   %global NQUANT;
   %global LNG_NQUANT;
   %global LNGROBUST_NQUANT;
   %global LNGROBUST_KRI_USED;
   %global NEEDED_INPUT;
   %global SCALE_TRANSFORM;
   %global DIST_NAME;
   %global PARAM_COUNT;
   %global ProbMeasure;
   %global COV_COUNT;
   %global NPARMS;
   %global SUMLOG; 
   %global CDF_TRANSFORM_FACTOR;
   %global BODY_DIST;
   %global TAIL_DIST;
   %global TAIL_PARAM_1;
   %global TAIL_PARAM_2;
   %global BODY_PARAM_COUNT;
   %global KRI_LOG_AVG;
   %global AVG_ADJ_GOF_T;     
   %global STATUS;      
   %global KRI_LOG;
   %global KRI_LIST;
   %global KRI_EFFECT;
   %global KRI_EFFECT_RETAIN;
   %global KRI_LOSS_EFFECT;
   %global KRI_RETAIN;
   %global KRI_AVG_EFFECT;
   %global Use_42_Scen_Calc;
   %global ORVP_STATUS;

   %_orv_insuranceSetGlobals();
   %orvu_delete_data(data = _ifs_DistNames);
%mend orva_init_global_macros;

%macro _orv_insuranceSetGlobals();
   %global InsuranceArray         ;   
   %global pInsurance             ;
   %global coverageData1          ;

   %global CoverageCount          ;
   %let    CoverageCount = 0      ;
 
   %global InsuranceContractCount    ;
   %let    InsuranceContractCount = 0;

   %global pInsuranceCount        ;
   %let    pInsuranceCount = 0    ;

   %global Insurance_UID_1        ;
   %global Insurance_Index_1      ;
   %global Deductible_1           ;
   %global The_Limit_1            ;
   %global Eff_Agg_Limit_1        ;
   %global Payout_Haircut_1       ;
   %global PolicyExpiryDate_1     ;

   %global Insurance_UID_2        ;
   %global Insurance_Index_2      ;
   %global Deductible_2           ;
   %global The_Limit_2            ;
   %global Eff_Agg_Limit_2        ;
   %global Payout_Haircut_2       ;
   %global PolicyExpiryDate_2     ;
%mend _orv_insuranceSetGlobals;
   
