/*************************************************************************************
 * Copyright (C), 2010
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME: orva_insurance_apply_basel.sas                                             
 * PURPOSE: Apply individual loss insurance recovery according to the BASEL rule.
 * SUPPORT: jariou
 *
 * PARAMETERS: NONE
 *
 * SUBORDINATE MACROS:
 *    _InsuranceApplyBasel(Level=);
 **************************************************************************************/
%macro orva_insurance_apply_basel();
   benefit_1 = 0;                    
   benefit_2 = 0;                    
   CoverageIndex = rand("TABLE", &pInsurance);

   /* In case the probabilities of coverage sum to less than 1 */
   if CoverageIndex LE &pInsuranceCount then do; 

      /* Check if there is primary Insurance coverage */
      if arInsurance_UID_1_[CoverageIndex] NE 0 then do; 
      %_orva_InsuranceApplyBasel(Level = 1);
         iloss = max(iloss - benefit1, 0);

         /* Check if there is secondary Insurance coverage */
         if arInsurance_UID_2_[CoverageIndex] NE 0 and iloss > 0 then do;
            %_orva_InsuranceApplyBasel(Level = 2);    
         end;    

         iLoss = max(iLoss - benefit_2, 0);   
      end; /* End of: Check if there is primary Insurance coverage */
   end; /* End of: In case the probabilities of coverage sum to less than 1 */
%mend orva_insurance_apply_basel;

%macro _orva_InsuranceApplyBasel(Level = );
   /* Apply insurance only if the policy is not expiring soon */
   if arPolicyExpiryDate_&Level._[CoverageIndex] - &simDate GE 90 then do;

      /*- handle aggregate limit when applicable -*/
      if arEff_Agg_Limit_&Level._[CoverageIndex] then do;  /*Level 1*/

         /*- if the accumulated benefits have not reached the aggregate limit, -*/
         /*- calculate the amount of insurance benefit -*/
         if sumInsurance < arEff_Agg_Limit_&Level._[CoverageIndex] then do; /*Level 2*/

            /*Just in case some Haircut is missing in the Insurance data */
            if arPayout_Haircut_&Level._[CoverageIndex] NE . then do;
               benefit_&Level. = arPayout_Haircut_&Level._[CoverageIndex]; 
               benefit_&Level. = benefit_&Level. * max(0, min(iloss - arDeductible_&Level._[CoverageIndex], 
                                                            arLimit_&Level._[CoverageIndex] - arDeductible_&Level._[CoverageIndex]
                                                            )
                                                      );
            end;
            else do;
               benefit_&Level. = max(0, min(  iloss - arDeductible_&Level._[CoverageIndex],
                                             arLimit_&Level._[CoverageIndex] - arDeductible_&Level._[CoverageIndex]
                                           )
                                    );
            end;

            sumInsurance = sumInsurance + benefit_&Level.;

            /*- check whether we passed the aggregate limit  -*/
            /*- this test will be true at most once  -*/
            if sumInsurance >= arEff_Agg_Limit_&Level._[CoverageIndex] then do; /*Level 3*/
               benefit_&Level. = benefit_&Level. + 
                                 arEff_Agg_Limit_&Level._[CoverageIndex] - 
                                 sumInsurance
                                 ;
               sumInsurance = arEff_Agg_Limit_&Level._[CoverageIndex];   
            end;/*- we have reached the aggregate limit, no more insurance benefit -*/
         end;    /*Level 2*/
      end; /*Level 1*/
      else do;
         benefit_&Level. = arPayout_Haircut_&Level._[CoverageIndex];
         benefit_&Level. = benefit_&Level. * max(0, min( iloss - arDeductible_&Level._[CoverageIndex],
                                                         arLimit_&Level._[CoverageIndex] - arDeductible_&Level._[CoverageIndex]
                                                        )
                                                 );
      end; /*Level 1*/

      Insurance_[arInsurance_Index_&Level._[CoverageIndex]] = benefit_&Level. + 
                                                              Insurance_[arInsurance_Index_&Level._[CoverageIndex]] 
                                                              ;
   end;
%mend _orva_InsuranceApplyBasel;
