/*************************************************************************************
 * Copyright (C), 2010
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME: orva_insurance_apply_plain.sas                                             
 * PURPOSE: Apply PLAIN individual loss insurance recovery.
 * SUPPORT: jariou
 *
 * PARAMETERS: NONE
 *
 * SUBORDINATE MACROS:
 *    _InsuranceApplyPlain(Level=);
 **************************************************************************************/
%macro orva_insurance_apply_plain();
   benefit_1 = 0;                    
   benefit_2 = 0;                    
   CoverageIndex = rand("TABLE", &pInsurance);

   /* In case the probabilities of coverage sum to less than 1 */
   if CoverageIndex LE &pInsuranceCount then do; 

      /* Check if there is primary Insurance coverage */
      if arInsurance_UID_1_[CoverageIndex] NE 0 then do; 
         %_orva_InsuranceApplyPlain(Level = 1);
         iloss = max(iloss - benefit1, 0);

         /* Check if there is secondary Insurance coverage */
         if arInsurance_UID_2_[CoverageIndex] NE 0 and iloss > 0 then do;
            %_orva_InsuranceApplyPlain(Level = 2);    
         end;    

         iLoss = max(iLoss - benefit_2, 0);   
      end; /* End of: Check if there is primary Insurance coverage */
   end; /* End of: In case the probabilities of coverage sum to less than 1 */
%mend orva_insurance_apply_plain;

%macro _orva_InsuranceApplyPlain(Level=);
   /*- handle aggregate limit when applicable -*/
   if arEff_Agg_Limit_&Level._[CoverageIndex] then do;  

      /*- if the accumulated benefits have not reached the aggregate limit, -*/
      /*- calculate the amount of insurance benefit -*/
      if sumInsurance < arEff_Agg_Limit_&Level._[CoverageIndex] then do; 
         benefit_&Level. = max(0, min(  iloss - arDeductible_&Level._[CoverageIndex],
                                       arLimit_&Level._[CoverageIndex] - arDeductible_&Level._[CoverageIndex]
                                     )
                              );
         sumInsurance = sumInsurance + benefit_&Level.;

         /*- check whether we passed the aggregate limit  -*/
         /*- this test will be true at most once  -*/
         if sumInsurance >= arEff_Agg_Limit_&Level._[CoverageIndex] then do; /*Level 3*/
            benefit_&Level. = benefit_&Level. + 
                              arEff_Agg_Limit_&Level._[CoverageIndex] - sumInsurance
                              ;
            sumInsurance = arEff_Agg_Limit_&Level._[CoverageIndex];   
         end;/*- we have reached the aggregate limit, no more insurance benefit -*/
      end;    
   end; 
   else do;
      benefit_&Level. = max(0, min( iloss - arDeductible_&Level._[CoverageIndex],
                                    arLimit_&Level._[CoverageIndex] - arDeductible_&Level._[CoverageIndex]
                                   )
                            );
   end; 

   Insurance_[arInsurance_Index_&Level._[CoverageIndex]] =  benefit_&Level. + 
                                                            Insurance_[arInsurance_Index_&Level._[CoverageIndex]] 
                                                            ;
%mend _orva_InsuranceApplyPlain;
