/*************************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *    NAME: orva_loss_sim_stream.sas                                             
 * PURPOSE: sub-macro of %orva_loss_sim() called once for each stream.            
 * SUPPORT: jariou
 *
 * PARAMETERS:
 *                startrep    = 1st rep number.
 *                nreps       = Number of reps for stream.
 *                streamNum   = Current stream number.
 *                seed        = Initial seed for stream.
 *
 * NOTES: Please see the accompanying doc for %orva_loss_sim().
 *        The only output of this macro are the datasets:
 *           _osimNaked_&streamNum 
 *           _osimInsured_&streamNum
 *
 * SUBORDINATE MACROS:
 *                      %_orva_RunSimulation(lookupTable=, applyKris=);
 **************************************************************************************/
%macro orva_loss_sim_stream(
                            startRep        = , 
                            nReps           = ,
                            streamNum       = ,
                            seed            = ,
                            isTotalCell     = ,
                            insuranceMethod = ,
                            finald          = work
                            );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #orva_loss_sim_stream(;
      %put .                      startRep        = &startRep.,; 
      %put .                      nReps           = &nReps.,;
      %put .                      streamNum       = &streamNum.,;
      %put .                      seed            = &seed.,;
      %put .                      isTotalCell     = &isTotalCell.,;
      %put .                      insuranceMethod = &insuranceMethod.,;
      %put .                      finald          = &finald.;
      %put .                      );
   %end;

   data 
        &finalD.._osimNaked_&streamNum (keep = 
                                               _rep_ 
                                               loss 
                                               TotalLoss  
                                               TotalLoss2  
                                               TotalCount  
                                               TotalCount2 
                                               )
         &finalD.._osimInsured_&streamNum(Keep = 
                                                 _rep_ 
                                                 loss 
                                                 &InsuranceArray
                                                 );
      retain 
            TotalLoss   0 
            TotalLoss2  0 
            TotalCount  0 
            TotalCount2 0
            ;
    
      if _N_ = 1 then do;
         %if "&AVERAGE_KRI_EFFECT." NE "1" %then %do;
            declare hash kriLookup(dataset: "_tmp_kri_", hashexp: 20);
            rc = kriLookup.definekey('Kri_Hash_Index');
            rc = kriLookup.definedata('KRI_SCALE_EFFECT');
            kriLookup.definedone();
         %end;

         declare hash freqLookup(dataset: "_loss_", hashexp: 20);
         rc = freqLookup.definekey('Freq_Hash_Index');
         rc = freqLookup.definedata('nloss');
         freqLookup.definedone();
      end;
 
      /* declare fake array to make sure insurance values are referenced
      otherwise, sas does not 'keep' them... */
      %if &InsuranceArray ne %str() %then %do;
         array _dummy[*] &InsuranceArray;
      %end;
    
      /* Initialize the random stream */
      call streaminit(&seed);

      /* Always create a main lookup table array */
      array mainLookupTable(&lookupSize) _TEMPORARY_ (&LOOKUP_QUANTILES);

      %if "&InsuranceArray" ne "" %then %do;
         array Insurance_[*] &InsuranceArray;

         %if &CoverageCount = 0 %then %do;
            array arInsurance_UID_1_[1]    _TEMPORARY_ ;      
            array arInsurance_Index_1_[1]  _TEMPORARY_;      
            array arDeductible_1_[1]       _TEMPORARY_ ;      
            array arLimit_1_[1]            _TEMPORARY_ ;                      
            array arEff_Agg_Limit_1_[1]    _TEMPORARY_ ;     
            array arPayout_Haircut_1_[1]   _TEMPORARY_ ;    
            array arPolicyExpiryDate_1_[1] _TEMPORARY_;    
            array arInsurance_UID_2_[1]    _TEMPORARY_ ;      
            array arInsurance_Index_2_[1]  _TEMPORARY_ ;      
            array arDeductible_2_[1]       _TEMPORARY_;            
            array arLimit_2_[1]            _TEMPORARY_;                      
            array arEff_Agg_Limit_2_[1]    _TEMPORARY_ ;              
            array arPayout_Haircut_2_[1]   _TEMPORARY_ ;    
            array arPolicyExpiryDate_2_[1] _TEMPORARY_ ;
            %let pInsurance = 0;    
         %end;
         %else %do;
            array arInsurance_UID_1_[&CoverageCount]    _TEMPORARY_ ( &Insurance_UID_1    );      
            array arInsurance_Index_1_[&CoverageCount]  _TEMPORARY_ ( &Insurance_Index_1  );      
            array arDeductible_1_[&CoverageCount]       _TEMPORARY_ ( &Deductible_1       );      
            array arLimit_1_[&CoverageCount]            _TEMPORARY_ ( &The_Limit_1        );                      
            array arEff_Agg_Limit_1_[&CoverageCount]    _TEMPORARY_ ( &Eff_Agg_Limit_1    );     
            array arPayout_Haircut_1_[&CoverageCount]   _TEMPORARY_ ( &Payout_Haircut_1   );    
            array arPolicyExpiryDate_1_[&CoverageCount] _TEMPORARY_ ( &PolicyExpiryDate_1 );    
            array arInsurance_UID_2_[&CoverageCount]    _TEMPORARY_ ( &Insurance_UID_2    );      
            array arInsurance_Index_2_[&CoverageCount]  _TEMPORARY_ ( &Insurance_Index_2  );      
            array arDeductible_2_[&CoverageCount]       _TEMPORARY_ ( &Deductible_2       );            
            array arLimit_2_[&CoverageCount]            _TEMPORARY_ ( &The_Limit_2        );                      
            array arEff_Agg_Limit_2_[&CoverageCount]    _TEMPORARY_ ( &Eff_Agg_Limit_2    );              
            array arPayout_Haircut_2_[&CoverageCount]   _TEMPORARY_ ( &Payout_Haircut_2   );    
            array arPolicyExpiryDate_2_[&CoverageCount] _TEMPORARY_ ( &PolicyExpiryDate_2 );    
         %end;
      %end;

      /*  Later, I will set loss to sumNaked, then output to outsimNaked, 
      then set loss to sumInsured and output to outsimInsured.    */

      /* TODO: */ 
      /* Could replace the set _loss_ point=??? */
      /* by an array written to a file before   */
      /* slope = (lookupTable[&LookupSize] - lookupTable[&LookupSize-1]); */

      /* Loop through replications */
      do curr_rep = &startRep to &startRep + &nreps - 1 ;
         Freq_Hash_Index = curr_rep;         
         rc    = freqLookup.find();

         %if "&InsuranceArray" ne "" %then %do;
            do ii = 1 to &InsuranceContractCount;
               Insurance_[ii] = 0;
            end;
         %end;    

         /*- simulate severities -*/
         sumNaked     = 0;
         sumInsured   = 0;
         sumInsurance = 0;

         TotalCount  = TotalCount  + nloss;
         TotalCount2 = TotalCount2 + nloss * nloss;

         /* Loop through individual losses */
         do i = 1 to nloss; 
            %if "&AVERAGE_KRI_EFFECT" NE "1" %then %do;
               %_orva_RunSimulation( 
                                    lookupTable = mainLookupTable, 
                                    applyKris   = Yes,
                                    insMethod   = &insuranceMethod
                                    );
            %end;
            %else %do;
               %_orva_RunSimulation( 
                                    lookupTable = mainLookupTable, 
                                    applyKris   = No,
                                    insMethod   = &insuranceMethod
                                    );
            %end;
         end; /* End of Loop through individual losses */

         _rep_  = curr_rep;
         loss   = sumNaked;                  
         output &finalD.._osimNaked_&streamNum;

         loss   = sumInsured;    
         output &finalD.._osimInsured_&streamNum;
      end; /* End of Loop through replications */
   run;
   %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg( &orvads, lostream_cant_sim_agg_loss_error, noquote, &streamNum));

      data &finalD.._osimNaked_&streamNum; 
         set _NULL_; 
      run;

      data &finalD.._ostatNaked_&streamNum; 
         set _NULL_; 
      run;
   %end;
%mend orva_loss_sim_stream;
                                               
%macro _orva_RunSimulation(
                           lookupTable = , 
                           applyKris   = , 
                           insMethod   = 
                           );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @_orva_RunSimulation(;
      %put .                     lookupTable = &lookupTable.,; 
      %put .                     applyKris   = &applyKris.,; 
      %put .                     insMethod   = &insMethod.;
      %put .                     );
   %end;

   uni       = rand("UNIFORM"); 
   LossIndex = (&LookupSize + 1) * (1 - uni**.4);
   FL_Index  = Floor(LossIndex);

   if FL_Index EQ 0 then do;
      iLoss = (1 - (1 - LossIndex)**2.5) * &lookupTable[1];
   end;
   else if FL_Index GE &LookupSize then do;
      iLoss = &lookupTable[&LookupSize] + 
               (LossIndex - &LookupSize) * 
               (&lookupTable[&LookupSize] - &lookupTable[&LookupSize - 1]);
   end;
   else do;
      iLoss = &lookupTable[FL_Index] + 
               (LossIndex - FL_Index) * (&lookupTable[FL_Index + 1] - &lookupTable[FL_Index]);
   end;

   /* TODO: */
   /* Speed up can be accomplished here by putting all of this in an array */
   /* Scale appropriately if KRI's are used */
   %if "&applyKris" EQ "Yes" %then %do;
      uChooseKriScale = 1 + int(&KRI_CHOICES * rand("UNIFORM"));
      Kri_Hash_Index = uChooseKriScale;         
      rc             = kriLookup.find();
      iLoss = iLoss * KRI_SCALE_EFFECT / &AVERAGE_KRI_EFFECT;
   %end;

   if iLoss > &MaxLoss then do;
      iLoss = &MaxLoss;
   end;

   TotalLoss  = TotalLoss + iLoss;
   TotalLoss2 = TotalLoss2 + iLoss * iLoss;
   sumNaked   = sumNaked + iloss;

   /*- apply insurance contract -*/
   %if "&InsuranceArray" ne "" %then %do;  
      %if (%upcase( &isTotalCell ) NE TRUE ) %then %do;
         %orva_insurance_apply_&insMethod.();
      %end;
   %end; 

   sumInsured = sumInsured + iloss;                
%mend _orva_RunSimulation;/* end of %RunSimulation */
