/************************************************************************************* 
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME: orva_sfitkri_dist.sas                                             
 *
 * PURPOSE: sub-macro of %orva_sfit_kri() called once for each distribution.            
 *
 * SUPPORT: jariou
 *
 * PARAMETERS:
 *          CELLDATA    = loss data for cell to fit.
 *          DIST        = Current distribution.
 *          NIDST       = Current distribution number.
 *          SUMMARYDATA =
 *    
 * NOTES: Please see the accompanying doc for %orva_sfit_kri()
 *        The only output of this macro are the datasets 
 *            _parmnlp_&ndist and _covmat_&ndist.
 *
 **************************************************************************************/
%macro orva_sfitkri_dist(
                           cellData    =, 
                           dist        =, 
                           ndist       =, 
                           summaryData =
                           );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;            
      %put #orva_sfitkri_dist(;
      %put .                   cellData    = &cellData.,; 
      %put .                   dist        = &dist.,;
      %put .                   ndist       = &ndist.,; 
      %put .                   summaryData = &summaryData.;
      %put .                   );
   %end;

   %let PARAM_COUNT = .;
   %let AVG_P_OBS   = -1; /* initializing */
   %orva_set_sev_dist( 
                        distIn  = &dist, 
                        distOut = dist
                        );

   /* reload macros for this distribution          */
   /* this also gives us the number of parameters  */
   /* in the model (in macro PARAM_COUNT)          */
   %orvd_&dist;

   data _NULL_;
      tmp = "&PARAM_COUNT";
      call symput("PARAM_COUNT", trim(left(tmp)));
   run;

   %let status = 0;
   %let reason = %sysfunc(sasmsg(&orvads, successful_note, noquote));
   %let nparms = 0;

   %do i = 1 %to  &MAX_SEV_PARAMS;
      %let param_&i._init = . ;
   %end;

   %InitFit&dist(
                  summaryData = &summaryData, 
                  KRI_FIT     = YES
                  );

   %if %sysevalf(&status NE 0,boolean) %then %do;
      %goto LOCAL_EXIT;
   %end;

   %if %sysevalf(&PARAM_COUNT + &COUNT_KRI > &obsCount,boolean) %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, sfdist_less_data_error, noquote));
      %let reason = %sysfunc(sasmsg(&orvads, sfdist_less_data_error, noquote));
      %let status = 1;
      %goto LOCAL_EXIT;
   %end;

   %let PROC_NLP_OPTIONS = %str(&BASE_PROC_NLP_OPTIONS.);

   /* Comment out this line to take out optimization for large datasets */
   %orv_optimize_fitting(withKRIs = YES);
   %orvu_delete_data( data = _cs_ );
  
   %if &status. eq 0 %then %do;
      %FitKRI&dist;
      %let nobs = 0;
      %let dsid = %sysfunc(open(_cs_, I));

      %if &dsid ^= 0 %then %do;
         %let nobs = %sysfunc(attrn(&dsid, nlobs));
         %let rc   = %sysfunc(close(&dsid));
      %end;

      %if &nobs eq 0 %then %do;
         %let status = 1;
         %let reason = %sysfunc(sasmsg(&orvads, sfdist_conv_fail_error, noquote));
      %end;

      %if &status. = 0 %then %do;
         data _null_;
            set _cs_;
            call symputX("status", status);
            call symputX("reason", reason);
         run;
      %end;
      %if &status. ne 0 %then %do;
         %let status = 1;
      %end;
   %end;

   %if &status = 0 %then %do;
      data _covmat_&ndist.;
         format Distribution &ORVAR_DISTRIBUTION_FMT. ;
         set _covmat_;

         array Params[*] Param_:;

         if RowName = "Param_0" then do;
               RowName = "Param_1";
         end;

         %if &SCALE_TRANSFORM = ID %then %do;
            Param_0 = Param_0 * exp(&The_Intercept); 

            if RowName = "Param_1" then do;
               do i = 1 to dim(Params);
                  Params[i] = Params[i] * exp(&The_Intercept);
               end;
            end;        
         %end;

         length PARAM_TYPE $10;

         if "&dist" = "MIXDIST" then do;
            distribution = "_GPD_&BODY_DIST";
         end;
         else do;
            distribution = "&dist";
         end;

         PARAM_TYPE     = "LOSS_MODEL";
         Rename Param_0 = Param_1;
      run;

      data _null_;
         set _logl_;

         if _N_ = 3 then do;
            call symputX("logl", put(nValue1, BEST20.));
            aic = -2 * (nValue1 - (&nparms + &COUNT_KRI));
            sbc = -2 * (nValue1 - (&nparms + &COUNT_KRI) * log(&NExcess) / 2);
            call symputX("aic", put(aic, BEST20.));
            call symputX("sbc", put(sbc, BEST20.));
         end;
      run;

      %let NLP_COV_COUNT = 0;
      %orvd_&dist;

      /*  Get estimated parameters from the NLP part  */
      data _Parmnlp_(keep = param_:);
         set _Parmnlp_ (where = (index(_type_, "PARMS") > 0));

         %if &SCALE_TRANSFORM = LOG %then %do;
            Param_0 = Param_0 + &The_Intercept; 
         %end;
         %else %do;
            Param_0 = Param_0 * exp(&The_Intercept); 
         %end;

         rename  Param_0 = Param_1;
      run;    
      %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, sfdist_cant_extract_paramest_error, noquote));
         %let reason = %sysfunc(sasmsg(&orvads, sfdist_cant_extract_paramest_error, noquote));
         %goto LOCAL_EXIT;
      %end;

      %let KRI_MODEL_NOI = ;
      /*  Format the final returned parameters dataset    */
      data _parmnlp_(drop = Intercept);
         format 
               distribution $12. 
               reason       &ORVAR_REASON_FMT 
               ;    
         set _parmnlp_ ;

         /* Create the KRI_MODEL_NOI macro variable */
         array parms[*] param_1_:;
         length KRI_MODEL $2000;
         KRI_MODEL =  "0"; 

         do i = 1 to dim(parms);
            KRI_MODEL = trim(KRI_MODEL) || " + " || substrn(vname(parms[i]),9) || " *" ||parms[i]; 
         end;

         call symput("KRI_MODEL_NOI", trim(KRI_MODEL));
         /* Done creating the KRI_MODEL_NOI macro variable */

         if "&dist" = "MIXDIST" then do;
            distribution = "_GPD_&BODY_DIST";
         end;
         else do;
            distribution = "&dist";
         end;

         KRI_Select_Method = "%upcase(&Sev_Selection_Method_Kris)";
         status            = &status;
         reason            = "&reason";
         logl              = &logl;
         aic               = &aic;
         sbc               = &sbc;
      run;
      %if &SYSERR ^= 0 & &SYSERR ^= 4 %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, sfdist_cant_prepare_ds_for_paramest_error, noquote));
         %let reason = %sysfunc(sasmsg(&orvads, sfdist_cant_prepare_ds_for_paramest_error, noquote));
         %goto LOCAL_EXIT;
      %end;

      %let DiffKriVectorCount = %str();
      %let Log_Scale          = %str();
      %let The_Prob           = %str();

      %orva_get_kri_log(data           = _parmnlp_);
      %orva_get_kri_mix(
                        lossData       = &cellData,
                        paramData      = _parmnlp_,
                        KRIs           = &KRI_LIST,
                        KriVectorCount = DiffKriVectorCount,
                        Scales         = Log_Scale,
                        Weights        = The_Prob
                        );

      /* Decide wether to use the adjusted testing threshold or not */
      %if &AVG_ADJ_GOF_T NE . %then %do; 
         %ORVA_GFit(
                     data         = &cellData, 
                     var          = AMOUNT, 
                     thresholdVar = Threshold, 
                     dist         = &dist, 
                     parmdata     = _parmnlp_, 
                     gofThreshold = &gofthreshold
                     );  
      %end;
      %else %do;
         %ORVA_GFit(
                     data         = &cellData, 
                     var          = AMOUNT, 
                     thresholdVar = Threshold, 
                     dist         = &dist, 
                     parmdata     = _parmnlp_, 
                     gofThreshold = 
                     );  
      %end;

      %orv_KriFindMoments(
                           parmdata = _parmnlp_, 
                           dist     = &dist
                           );    

      /* add _PARMNLP_ data to every row in &celldata */
      data _Data_plus_params_(compress = binary); 
         set &celldata;

         if _N_ eq 1 then do;
            set _PARMNLP_;
         end;
      run;

      /* calculate the avg_p_obs */
      data _NULL_ (drop = p_obs_sum p_obs_count);
         set _Data_plus_params_ end = last;
         retain 
               p_obs_sum   0 
               p_obs_count 0
               ;

         if THRESHOLD = 0 then do;
            p_obs = 0;
         end;
         else do;
            %if "&SCALE_TRANSFORM" = "LOG" %then %do;
               Param_1 = Param_1 + &KRI_LOG; 
            %end;
            %else %do;
               Param_1 = Param_1 * exp(&KRI_LOG); 
            %end;

            %VCdf&dist(THRESHOLD, p_obs);
         end;

         if ^missing(p_obs) then do;
            p_obs_sum   = p_obs_sum + p_obs;
            p_obs_count = p_obs_count + 1;
         end;

         if last then do;
            call symputX("avg_p_obs", put( 1 - p_obs_sum / p_obs_count, BEST20.));
         end;
      run;
   %end;
   %else %do;
%LOCAL_EXIT:
      data _parmnlp_;
         format distribution $12. reason &ORVAR_REASON_FMT ;    

         if "&dist" = "MIXDIST" then do;
            distribution = "_GPD_&BODY_DIST";
         end;
         else do;
            distribution = "&dist";
         end;

         status = &status;
         reason = "&reason";
      run;   

      data _covmat_&ndist.; 
         set _NULL_; 
      run;
   %end;

   data _parmnlp_&ndist;
      set _parmnlp_;
      obsCount      = &obsCount;
      ParamCount    = &PARAM_COUNT;
      COUNT_KRI     = &COUNT_KRI;
      avg_p_obs     = &AVG_P_OBS;
   run;

   /***** cleanup temporary datasets *****/
   %orvu_delete_data( data =
                              _parmnlp_  
                              _NLP_COV_  
                              _Data_plus_params_
                              );
%mend orva_sfitkri_dist;

/*- find moments for fitted distribution -*/
%macro orv_KriFindMoments(
                           parmdata =, 
                           dist     = 
                           );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;            
      %put @orv_KriFindMoments(;
      %put .                    parmdata = &parmdata.,; 
      %put .                    dist     = &dist.;
      %put .                    );
   %end;

   %orvd_&dist;

   data &parmdata;
      set &parmdata;
      array Scales[&DiffKriVectorCount] _TEMPORARY_ (&Log_Scale);
      array Probs[&DiffKriVectorCount] _TEMPORARY_ (&The_Prob);

      tmp_m1 = 0;
      tmp_m2 = 0;
      tmp_m3 = 0;
      tmp_m4 = 0;

      tmp_Param_1 = Param_1;

      do i = 1 to &DiffKriVectorCount;
         %if "&SCALE_TRANSFORM" = "LOG" %then %do;
            Param_1 = tmp_Param_1 + Scales[i]; 
         %end;
         %else %do;
            Param_1 = tmp_Param_1 * exp(Scales[i]); 
         %end;

         %Moments&dist;

         tmp_m1 = tmp_m1 + m1 * Probs[i];
         tmp_m2 = tmp_m2 + m2 * Probs[i];
         tmp_m3 = tmp_m3 + m3 * Probs[i];
         tmp_m4 = tmp_m4 + m4 * Probs[i];
      end;

      mean     = tmp_m1;
      std      = sqrt(tmp_m2 - tmp_m1 * tmp_m1);

      if not missing(std) then do;
         skewness = (tmp_m3 - 3 * tmp_m1 * tmp_m2 + 2 * (tmp_m1**3))/(std**3);
         kurtosis = (tmp_m4 - 4 * tmp_m1 * tmp_m3 + 6 * (tmp_m1**2) * tmp_m2 - 3 * (tmp_m1**4))/(std**4);
      end;
      else do;
         skewness = .;
         kurtosis = .;
      end;

      Param_1 = tmp_Param_1;

      if mean = . then do;
         status = 1;
         reason = "%sysfunc(sasmsg(&orvads, sfdist_no_mean_error, noquote))";
      end;
      else if std = . then do;
         status = 2;
         reason = "%sysfunc(sasmsg(&orvads, sfdist_no_variance_error, noquote))";
      end;

      /* Insuring that the drop is not going to cause warnings. */
      _lm_1  = .;
      _lng_1 = .;
      _par_1 = .;

      drop 
           tmp_m1 
           tmp_m2 
           tmp_m3 
           tmp_m4 
           m1 
           m2 
           m3 
           m4 
           tmp_Param_1 
           _lm_: 
           _lng_: 
           _par_:
           ;
   run;
%mend orv_KriFindMoments;
