/******************************************************************************
 * Copyright (C), 2006
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *   Title: orvd_mixdist.sas
 *   
 *   -Provides support for the mixture of a distributions
 *
 *
 *  The macro %ORVD_MIXDIST is at the end of this file. It is a empty macro that takes
 *  no parameters and returns no feedback. In normal usage, this macro needs to be 
 *  called before trying to call any other macros depending on the fragments in this file.
 *
 *  Calling the %ORVD_LOGN macro makes the SAS system parse and compile the whole file 
 *  so that the other macro fragment then become defined and available for other 
 *  macros to call.
 *
 *  Of course this is only guaranteed to work if there is no syntax error in any of 
 *  the fragments. 
 *
 ******************************************************************************/

/*****************************************************************************************
 *   What it does  :  Evaluate the cumulative distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the CDF is evaluated
 *        valueHolder = name of variable inton which result is stored
 *****************************************************************************************/
/*The Cumulative Distribution Function (using macro variables as parameters)*/
%macro CdfMixDist(x,returnVal);
   %if &SCALE_TRANSFORM = ID %then %do;
      Eff_Param_xLimit = &Param_xLimit * &Param_1;
      Eff_Tail_Param_1 = &&&TAIL_PARAM_1 * &Param_1;
   %end;
   %else %do;
      Eff_Param_xLimit = &Param_xLimit * exp(&Param_1);
      Eff_Tail_Param_1 = &&&TAIL_PARAM_1 * exp(&Param_1);
   %end;

   %Cdf&Body_Dist(Eff_Param_xLimit, Param_pLimit)

   if &x LE Eff_Param_xLimit then do;
      %Cdf&Body_Dist(&x, &returnVal)
      &returnVal  = &returnVal * &Param_pnLimit/Param_pLimit;    
   end;
   else do;
      if &&&TAIL_PARAM_2 LE 0.00000000001 then do;
         &returnVal = &Param_pnLimit + (1-&Param_pnLimit)*(1-exp(-(&x-Eff_Param_xLimit)/Eff_Tail_Param_1));
      end;
      else do;
         &returnVal = &Param_pnLimit + (1-&Param_pnLimit)*(1-(1 + &&&TAIL_PARAM_2*(&x - Eff_Param_xLimit)/Eff_Tail_Param_1)**(-1/&&&TAIL_PARAM_2));
      end;
   end;
%mend CdfMixDist;

/*The Cumulative Distribution Function (using dataset variables as parameters)*/
%macro VCdfMixDist(x,returnVal);
   %if &SCALE_TRANSFORM = ID %then %do;
      Eff_Param_xLimit = Param_xLimit * Param_1;
      Eff_Tail_Param_1 = &TAIL_PARAM_1 * Param_1;
   %end;
   %else %do;
      Eff_Param_xLimit = Param_xLimit * exp(Param_1);
      Eff_Tail_Param_1 = &TAIL_PARAM_1 * exp(Param_1);
   %end;

   %VCdf&Body_Dist(Eff_Param_xLimit, &returnVal._Param_pLimit)
   
   if &x LE Eff_Param_xLimit then do;
        %VCdf&Body_Dist(&x,&returnVal)
        &returnVal  = &returnVal * Param_pnLimit/&returnVal._Param_pLimit;    
    end;
    else do;
        if &TAIL_PARAM_2 LE 0.00000000001 then do;
            &returnVal = Param_pnLimit + (1-Param_pnLimit)*(1-exp(-(&x - Eff_Param_xLimit)/Eff_Tail_Param_1));
        end;
        else do;
            &returnVal = Param_pnLimit + (1-Param_pnLimit)*(1-(1 + &TAIL_PARAM_2*(&x - Eff_Param_xLimit)/Eff_Tail_Param_1)**(-1/&TAIL_PARAM_2));
        end;
    end;
%mend VCdfMixDist;

/*****************************************************************************************
 *   What it does  :  Evaluate the density distribution function (CDF)
 *
 *   Parameters    :
 *              x     = value at which the density is evaluated
 *        valueHolder = name of variable inton which result is stored
 *****************************************************************************************/
/*The Probability Distribution Function (using macro variables as parameters)*/
%macro PdfMixDist(x,returnVal);
   %if &SCALE_TRANSFORM = ID %then %do;
      Eff_Param_xLimit = &Param_xLimit * &Param_1;
      Eff_Tail_Param_1 = &&&TAIL_PARAM_1 * &Param_1;
   %end;
   %else %do;
      Eff_Param_xLimit = &Param_xLimit * exp(&Param_1);
      Eff_Tail_Param_1 = &&&TAIL_PARAM_1 * exp(&Param_1);
   %end;

   %Cdf&Body_Dist(Eff_Param_xLimit, Param_pLimit)
   
   if &x LE Eff_Param_xLimit then do;
        %Pdf&Body_Dist(&x,&returnVal)
        &returnVal = &returnVal * &Param_pnLimit/Param_pLimit;    
    end;
    else do;
        if &&&TAIL_PARAM_2 LE 0.00000000001 then do;
            &returnVal = (1-&Param_pnLimit) * exp(-(&x - Eff_Param_xLimit)/Eff_Tail_Param_1 )/Eff_Tail_Param_1 ;
        end;
        else do;
            &returnVal = (1-&Param_pnLimit) * ((1 + (&&&TAIL_PARAM_2)*(&x - Eff_Param_xLimit)/Eff_Tail_Param_1 )**(-1/&&&TAIL_PARAM_2 -1))/Eff_Tail_Param_1 ;
        end;
    end;
%mend PdfMixDist;
/*The Probability Distribution Function (using dataset variables as parameters)*/
%macro VPdfMixDist(x,returnVal);
   %if &SCALE_TRANSFORM = ID %then %do;
      Eff_Param_xLimit = Param_xLimit * Param_1;
      Eff_Tail_Param_1 = &TAIL_PARAM_1 * Param_1;
   %end;
   %else %do;
      Eff_Param_xLimit = Param_xLimit * exp(Param_1);
      Eff_Tail_Param_1 = &TAIL_PARAM_1 * exp(Param_1);
   %end;

   %VCdf&Body_Dist(Eff_Param_xLimit, &returnVal.Param_pLimit)

   if &x LE Eff_Param_xLimit then do;
        %VPdf&Body_Dist(&x,&returnVal)
        &returnVal = &returnVal * Param_pnLimit/&returnVal.Param_pLimit;    
    end;
    else do;
        if &TAIL_PARAM_2 LE 0.00000000001 then do;
            &returnVal = (1-Param_pnLimit) * exp(-(&x-Eff_Param_xLimit)/Eff_TAIL_PARAM_1)/Eff_TAIL_PARAM_1;
        end;
        else do;
            &returnVal = (1-Param_pnLimit)*((1 + (&TAIL_PARAM_2)*(&x-Eff_Param_xLimit)/Eff_TAIL_PARAM_1)**(-1/&TAIL_PARAM_2 -1))/Eff_TAIL_PARAM_1;
        end;
    end;
%mend VPdfMixDist;

/*****************************************************************************************
 *   What it does  :  Generates a dataset named _temp_ containing probability values
 *                    and their corresponding quantiles
 *
 *   No Parameters    
 *****************************************************************************************/
/* The quantile function is probably not working correctly presently. Still needs work*/
%macro QuantMixDist(scaleFactor=1);
   data _NULL_;
      %if &SCALE_TRANSFORM = ID %then %do;
         Eff_Param_xLimit = &Param_xLimit * &Param_1;
         Eff_Tail_Param_1 = &&&TAIL_PARAM_1 * &Param_1;
      %end;
      %else %do;
         Eff_Param_xLimit = &Param_xLimit * exp(&Param_1);
         Eff_Tail_Param_1 = &&&TAIL_PARAM_1 * exp(&Param_1);
      %end;

      %Cdf&Body_Dist(Eff_Param_xLimit ,tmp); 
      call symput("Param_pLimit", put(tmp,BEST20.)); 

      call symput("Eff_Param_xLimit", put(Eff_Param_xLimit,BEST20.) );
      call symput("Eff_Tail_Param_1", put(Eff_Tail_Param_1,BEST20.));
   run;

   %Quant&Body_Dist(scaleFactor = (&Param_pnLimit/&Param_pLimit))

  
    %if %sysevalf(&&&TAIL_PARAM_2 > 0, boolean) %then %do;
        %let quantTail = &Eff_Tail_Param_1 *((1/(1-tmp_cdf))**(&&&TAIL_PARAM_2)-1)/(&&&TAIL_PARAM_2);
    %end;
    %else %do;
        %let quantTail = - &Eff_Tail_Param_1 * log(1-tmp_cdf);
    %end;

    data _temp_;
      set _temp_ ;
   
      if (cdf > &Param_pnLimit ) then do;
         tmp_cdf = (cdf - &Param_pnLimit)/(1 - &Param_pnLimit);
         quant = &Eff_Param_xLimit + (&quantTail);
      end;

      drop tmp_cdf;
    run;

%mend QuantMixDist;

/*****************************************************************************************
 *   What it does  :  Calculates derivatives of the CDF function wrt parameters
 *
 *   Parameters    :
 *        scaleFactor = possible scaling for mixed distribution
 *        indata      = dataset of quantile points at which derivative is evaluated
 *        outdata     = dataset with derivatives 
 *****************************************************************************************/
%macro CdfDerMixDist(scaleFactor=1, indata=, outdata=);
   /* Data set for Body Distribution */
   data _temp_0;
   retain param_plimit Eff_Param_xLimit;

   %if &SCALE_TRANSFORM = ID %then %do;
      Eff_Param_xLimit = &Param_xLimit * &Param_1;
   %end;
   %else %do;
      Eff_Param_xLimit = &Param_xLimit * exp(&Param_1);
   %end;

   %Cdf&Body_Dist(Eff_Param_xLimit, Param_pLimit);
   set &indata.; 
      if quant > Eff_Param_xLimit then do;
         delete;
      end;
   run;

   /* Derivatives for the Body parameters */
   %CdfDer&Body_Dist.(scalefactor=&Param_pnLimit/Param_pLimit, indata=_temp_0, outdata=_Body_Derivs);

   /* Now do the Tail */
   data _temp_1;
   retain param_plimit Eff_Param_xLimit Eff_Tail_Param_1;

   %if &SCALE_TRANSFORM = ID %then %do;
      Eff_Param_xLimit = &Param_xLimit * &Param_1;
      Eff_Tail_Param_1 = &&&TAIL_PARAM_1 * &Param_1;
   %end;
   %else %do;
      Eff_Param_xLimit = &Param_xLimit * exp(&Param_1);
      Eff_Tail_Param_1 = &&&TAIL_PARAM_1 * exp(&Param_1);
   %end;

   %Cdf&Body_Dist(Eff_Param_xLimit, Param_pLimit);
   set &indata.; 
      if quant <= Eff_Param_xLimit then do;
         delete;
      end;

   if &&&TAIL_PARAM_2 LE 0.00000000001 then do;
      cdf = &Param_pnLimit + (1-&Param_pnLimit)*(1-exp(-(quant-Eff_Param_xLimit)/Eff_Tail_Param_1));
   end;
   else do;
      cdf = &Param_pnLimit + (1-&Param_pnLimit)*(1-(1 + &&&TAIL_PARAM_2*(quant - Eff_Param_xLimit)/Eff_Tail_Param_1)**(-1/&&&TAIL_PARAM_2));
   end;
   cdf1 = cdf;
      cdf2 = cdf;
   run;

   proc model data=_temp_1 noprint;
      dependent 
         cdf 
         cdf1;
      retain 
         param_1 &param_1 
         &tail_param_1 &&&TAIL_PARAM_1 
         &tail_param_2 &&&TAIL_PARAM_2;

      %if &SCALE_TRANSFORM = ID %then %do;
         Eff_Param_xLimit = &Param_xLimit * Param_1;
         Eff_Tail_Param_1 = &TAIL_PARAM_1 * Param_1;
      %end;
      %else %do;
         Eff_Param_xLimit = &Param_xLimit * exp(Param_1);
         Eff_Tail_Param_1 = &TAIL_PARAM_1 * exp(Param_1);
      %end;

      if &TAIL_PARAM_2 LE 0.00000000001 then do;
         cdf = &Param_pnLimit + (1-&Param_pnLimit)*(1-exp(-(quant-Eff_Param_xLimit)/Eff_Tail_Param_1));
      end;
      else do;
         cdf = &Param_pnLimit + (1-&Param_pnLimit)*(1-(1 + &TAIL_PARAM_2*(quant - Eff_Param_xLimit)/Eff_Tail_Param_1)**(-1/&TAIL_PARAM_2));
      end;
      cdf1=pred.cdf;

      %let param_number1 =%scan(&TAIL_PARAM_1,2,'_');
      %let param_number2 =%scan(&TAIL_PARAM_2,2,'_');

      DParam_&param_number1 = getder(pred.cdf,&tail_param_1);
      DParam_&param_number2 = getder(pred.cdf1,&tail_param_2);

      outvars 
         DParam_&param_number1 
         DParam_&param_number2
         quant;
      solve 
         Param_&param_number1 
         Param_&param_number2 /
      out=
         _Tail_Derivs
         (keep=
            cdf 
            quant 
            DParam_&param_number1 
            DParam_&param_number2 
         );
quit;

data &outdata;
   set _Body_Derivs _Tail_Derivs;
run;

/* Replace Missing values with 0 */
   proc stdize data=&outdata. 
      reponly 
      missing=0 
      out=&outdata.;
   var _numeric_;
run;
%mend CdfDerMixDist;

/*****************************************************************************************
 *   What it does  :  Simulates a random value
 *
 *   No Parameters    :  
 *****************************************************************************************/
%macro SimMixDist(loss);
   p=rand("UNIFORM");

   %if &SCALE_TRANSFORM = ID %then %do;
      Eff_Param_xLimit = Param_xLimit * Param_1;
      Eff_Tail_Param_1 = &TAIL_PARAM_1 * Param_1;
   %end;
   %else %do;
      Eff_Param_xLimit = Param_xLimit * exp(Param_1);
      Eff_Tail_Param_1 = &TAIL_PARAM_1 * exp(Param_1);
   %end;

   if p LE Param_pnLimit then do;
      _body_loop_count = 0;
      do until(&loss LE Eff_Param_xLimit);
            _body_loop_count = _body_loop_count + 1;
            if _body_loop_count < 100 then do;
               %Sim&Body_Dist(&loss)
            end;
            else do;
               &loss = rand("UNIFORM") * Eff_Param_xLimit;
            end;
      end;
      drop _body_loop_count;
   end;
   else do;
      if &TAIL_PARAM_2 > 0 then do;
            &loss = Eff_Param_xLimit  + (((1 / rand("UNIFORM"))**(&TAIL_PARAM_2) - 1) * &TAIL_PARAM_1 )/&TAIL_PARAM_2;        
      end;
      else do;
            &loss = Eff_Param_xLimit  - Eff_Tail_Param_1 * log(rand("UNIFORM"));
      end;

   end;
%mend SimMixDist;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro InitFitMixDist(summaryData = , KRI_FIT = NO);
    %InitFit&Body_Dist(summaryData = &summaryData, KRI_Fit = &KRI_FIT);
    %let nparms = &PARAM_COUNT;
    %put **********************************;
    %put Distribution    =  _GPD_&BODY_DIST;
    %put Observations    =  &obsCount;
    %put Param_1_init    =  &Param_1_init;

    %do j=2 %to &BODY_PARAM_COUNT;
         %put Param_&j._init = &&&Param_&j._init;
    %end;
    %put **********************************;
   
    %let &TAIL_PARAM_1._init=.;
    %let &TAIL_PARAM_2._init=.;

   /* BEGIN Does the fit use KRI's?  */
   %if %upcase(&KRI_FIT) EQ YES %then %do;
      %let _Loss_Data_    = _LogWorkdata_KRI_;
      %let _Loss_Amount_  = KRI_ADJ_AMOUNT;
   %end; 
   %else %do; 
      %let _Loss_Data_    = _LogWorkdata_;
      %let _Loss_Amount_  = AMOUNT;
   %end;
   /* END Does the fit use KRI's?    */

   %if "%upcase(&TailBoundaryChoice)" NE "TAILBOUNDARYPERCENT" %then %do;
      %let TailCount = 0;

      data _NULL_;
         set   &_Loss_Data_
               nobs = TotalCount 
               end  = TheEnd
               ;
         where AMOUNT <= &TailBoundaryValue;

         retain BodyCount 0;
         BodyCount = BodyCount + 1;

         if TheEnd then do;
            call symputX("TotalCount", TotalCount);   
            call symputX("BodyCount", BodyCount);
            call symputX("TailCount", TotalCount - BodyCount);
            call symputX("Param_pnLimit",put(BodyCount/TotalCount, BEST20.));
         end;
      run;

      %let TailBoundaryPercent = %sysevalf(&Param_pnLimit * 100);

      %if %sysevalf(&TailCount < 2, boolean) %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvdds, mixd_no_obs_above_break_error, noquote));
         %let status=1;
         %let reason=%sysfunc(sasmsg(&orvdds, mixd_no_obs_above_break_error, noquote));
         %goto LOCAL_DONE;
      %end;
   %end;
   %else %do;
      %let Param_pnLimit = %sysevalf(&TailBoundaryPercent/100);
   %end;
   


   /* At this point, the Param_pnLimit macro is defined */
   /* Get the right Param_xLimit                        */
   proc univariate data = &_Loss_Data_ PCTLDEF = 4 noprint;
      output 
            out     = __tmp__ 
            pctlpts = &TailBoundaryPercent 
            pctlpre = PCTL
            ;
      var &_Loss_Amount_;
   quit;

   proc transpose data=__tmp__ out =__tmp__;
   run;

   data _null_;
      set __tmp__;
      call symputx("Param_xLimit",put(COL1, BEST20.));
   run;
   /* Now Param_xLimit is defined in the right space */
   /* i.e in the AMOUNT space if no KRI's are used   */
   /*     in the AMOUNT/KRI_SCALE if KRI's are used  */

   /* BEGIN - Get initial values for tail parameters */
   proc summary data = &_Loss_Data_ noprint;
      var &_Loss_Amount_;
      where &_Loss_Amount_ > &Param_xLimit; 

      output 
            out  = _tail_stat_(keep=sample_mean2 sample_variance2 sample_size2)
            n    = sample_size2
            mean = sample_mean2 
            var  = sample_variance2 
            ; 
   run;

   %let TailStatCount = %str();
   %orvu_getObsCount(_tail_stat_,TailStatCount)

   %if %sysevalf(&TailStatCount = 0, boolean) %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvdds, mixd_no_obs_above_break_error, noquote));
      %let status=1;
      %let reason=%sysfunc(sasmsg(&orvdds, mixd_no_obs_above_break_error, noquote));
      %goto LOCAL_DONE;
   %end;

   proc transpose 
      data = _tail_stat_ 
      out  = _tail_stat_;
   run;
        
   data _null_;
      set _tail_stat_;
      call symputX(_name_, put(col1, BEST20.));
   run;
   %if %sysevalf(&sample_size2 < 2, boolean) %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvdds, mixd_no_obs_above_break_error, noquote));
      %let status=1;
      %let reason=%sysfunc(sasmsg(&orvdds, mixd_no_obs_above_break_error, noquote));
      %goto LOCAL_DONE;
   %end;
    
   data _null_;
      if &sample_variance2 > &sample_mean2**2 then do;
         alpha = (&sample_variance2 - &sample_mean2**2)/2 / &sample_variance2 ;
      end;
      else do;
         alpha = .49;
      end;

      theta = &sample_mean2*(1 - alpha);
      call symputX("&Tail_Param_2._init", put(alpha, BEST20.));

      %if &SCALE_TRANSFORM = ID %then %do;
         call symputX("&Tail_Param_1._init", put(theta/&Param_1_init, BEST20.));
      %end;
      %else %do;
         call symputX("&Tail_Param_1._init", put(theta/exp(&Param_1_init.), BEST20.));
      %end;
   run;
   /* END - Get initial values for tail parameters */

%LOCAL_DONE:
 
%mend InitFitMixDist;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitMixDist;    
   /* This piece of code increases Param_pnLimit if the probability of observability is provided     */
   /* New_pnLimit = Old_pnLimit + (1-Old_pnLimit)*p0 where p0 is the probability below the threshold */
   %if %sysevalf(&THRESHOLDPERCENT >= 0, BOOLEAN) %then %do;
      %let Param_pnLimit = %sysevalf(&Param_pnLimit + (1-&Param_pnLimit)* &THRESHOLDPERCENT); 
   %end;

    ods listing close;
    ods output 
               IterStop          =  _logl_
               ConvergenceStatus =  _cs_ 
               covmat            =  _covmat_
               ;  

   proc nlp 
            data   = _LogWorkdata_  
            outest = _parmnlp_ 
            &PROC_NLP_OPTIONS.
            ;

      /* The list of parameters */
      parms 
         Param_1 = &Param_1_init
         %do j=2 %to &BODY_PARAM_COUNT;
             ,Param_&j = &&&Param_&j._init
         %end;
         ,&Tail_Param_1 = &&&Tail_Param_1._init
            ,&Tail_Param_2 = &&&Tail_Param_2._init
         ;

      /* The bounds on parameters */
        bounds 
               %if &SCALE_TRANSFORM = ID %then %do;
                  Param_1 > 0 
               %end;
               %Param_Bounds&BODY_DIST()
               ,&TAIL_PARAM_1 > 0
               ,&TAIL_PARAM_2 > 0
               ; 

      /* The optimization part */
      max logf;

      _DELTA_       = 1.0e-6;
      Param_xLimit  = &PARAM_xLimit;
      Param_pnLimit = &PARAM_pnLimit;

      if THRESHOLD = 0 then do;
         lsdf = 0;
      end;
      else do;
         CdfOfThreshold = Cdf__GPD_&body_Dist(THRESHOLD, Param_1 &&_GPD_&body_Dist._Param_List_Minus_1);

         /* Handle the probability of observability */
         if &THRESHOLDPERCENT >= 0 then do;
            lsdf =  &THRESHOLDPERCENT/(1 - &THRESHOLDPERCENT) * 
                    log(CdfOfThreshold + 1E-30);
         end; 
         else do;
            lsdf = -log(1 - CdfOfThreshold + 1E-30);
         end;
      end;

      PdfOfLoss = Pdf__GPD_&body_Dist(amount,Param_1 &&_GPD_&body_Dist._Param_List_Minus_1);
        
      /* Combine all these pieces together     */
      /* to form the complete log-likelihood   */
      logf = log(PdfOfLoss + 1E-30) + lsdf;
   quit;
   ods listing;

    %if %sysfunc(exist(_cs_)) %then %do;
       data _null_;
           set _cs_;
           call symputx("status",status);
       run;

      %if &Status = 0 %then %do;
         data _parmnlp_;
            length TailBoundaryChoice $32;
            set _parmnlp_;
            where (index(_type_,"PARMS")>0);
            Param_pnLimit = &Param_pnLimit;

            if &Tail_Param_2 LE 0.0000000001 then do;
               &Tail_Param_2 = 0; 
            end;

            %if &SCALE_TRANSFORM = ID %then %do;
               Param_xLimit  = &Param_xLimit/Param_1;
            %end;
            %else %do;
               Param_xLimit  = &Param_xLimit/exp(Param_1);
            %end;
         run;
      %end;
      %else %do;
         %let status = 1;
         %let reason = %sysfunc(sasmsg(&orvdds,log_opt_problem_no_fit_error, noquote));
      %end;
    %end;
    %else %do;
        %let status = 1;
        %let reason = %sysfunc(sasmsg(&orvdds, log_body_fit_fail_error, noquote));
    %end;
%LOCAL_DONE:
%mend FitMixDist;

/*****************************************************************************************
 *   What it does  :  Actually does the minimization using PROC NLP
 *
 *   No Parameters    
 *****************************************************************************************/
%macro FitKriMixDist;
   /* This piece of code increases Param_pnLimit if the probability of observability is provided   */
   /* New_pnLimit = Old_pnLimit + (1-Old_pnLimit)*p0 where p0 is the probability of being hidden   */
   %if %sysevalf(&THRESHOLDPERCENT >= 0, BOOLEAN) %then %do;
      %let Param_pnLimit = %sysevalf(&Param_pnLimit + (1-&Param_pnLimit)* &THRESHOLDPERCENT); 
   %end;

   ods listing close;
   ods output 
               IterStop           =  _logl_
               ConvergenceStatus  =  _cs_ 
               covmat             =  _covmat_
               ;   

   proc nlp 
            data   = _LogWorkdata_KRI_(drop= PARAM_1:) 
            outest = _parmnlp_ 
            &PROC_NLP_OPTIONS.
            ;

      /* The list of parameters */     
      parms
            Param_0 = &Param_1_init
            %do j=2 %to &BODY_PARAM_COUNT;
               ,Param_&j = &&Param_&j._init
            %end;
            ,&Tail_Param_1 = &&&Tail_Param_1._init
            ,&Tail_Param_2 = &&&Tail_Param_2._init
            ,&KRI_INIT_PARM
            ;
            
      /* The bounds on parameters */      
      bounds 
            %if &SCALE_TRANSFORM = ID %then %do;
               Param_0 > 0 
            %end;
            %Param_Bounds&BODY_DIST()
            ,&Tail_Param_1 > 0
            ,&Tail_Param_2 > 0
            ; 

      max logf;
      _DELTA_ = 1.0e-6;

      %if &SCALE_TRANSFORM = LOG %then %do;
         Param_1       = Safe_ID(Param_0 + &KRI_MODEL_YES_INTERCEPT); 
         Param_xLimit  = &PARAM_xLimit * exp(Param_1);
      %end;
      %else %do;
         LogParam_1    = Safe_ID(log(Param_0) + &KRI_MODEL_YES_INTERCEPT );
         Param_1       = exp(LogParam_1); 
         Param_xLimit  = &PARAM_xLimit * Param_1;
      %end;

      Param_pnLimit = &PARAM_pnLimit;

      if THRESHOLD = 0 then do;
         lsdf = 0;
      end;
      else do;
         CdfOfThreshold = Cdf__GPD_&body_Dist(THRESHOLD, Param_1 &&_GPD_&body_Dist._Param_List_Minus_1);

         /* Handle the probability of observability */
         if &THRESHOLDPERCENT >= 0 then do;
            lsdf =  &THRESHOLDPERCENT/(1 - &THRESHOLDPERCENT) * 
                    log(CdfOfThreshold + 1E-30);
         end; 
         else do;
            lsdf = -log(1 - CdfOfThreshold + 1E-30);
         end;
      end;

      PdfOfLoss = Pdf__GPD_&body_Dist(amount,Param_1 &&_GPD_&body_Dist._Param_List_Minus_1);
      logf = log(PdfOfLoss + 1E-30) + lsdf;
    quit;
    ods listing;

   %if %sysfunc(exist(_cs_)) %then %do;
      data _null_;
         set _cs_;
         call symputx("status",status);
      run;

      %if &status = 0 %then %do;
         data _parmnlp_;
            set _parmnlp_ ;
            where (index(_type_,"PARMS")>0);
            Param_pnLimit = &Param_pnLimit;
            Param_xLimit  = &Param_xLimit;
         run;
      %end;
      %else %do;
         %let status = 1;
         %let reason = %sysfunc(sasmsg(&orvdds,log_opt_problem_no_fit_error, noquote));
      %end;
   %end;
   %else %do;
      %let status = 1;
      %let reason = %sysfunc(sasmsg(&orvdds, mixd_fit_fail_error, noquote));
   %end;
%LOCAL_DONE:
%mend FitKriMixDist;

/*- actually, the moments of logarithm of losses -*/
%macro MomentsMixDist;
   %if &SCALE_TRANSFORM = ID %then %do;
      _lm_t_Param_1 = &Tail_Param_1 * Param_1;
      _lm_tmp_xLimit = Param_xLimit * Param_1;
   %end;
   %else %do;
      _lm_t_Param_1 = &Tail_Param_1 * exp(Param_1);
      _lm_tmp_xLimit = Param_xLimit * exp(Param_1);
   %end;

   %VCDF&BODY_DIST(_lm_tmp_xLimit, _lm__tmp);
 
   /* Part of the moment due to the body distribution     */
   /* Need to use limited expectation of lognormal        */
   /* See formula in "Loss Models" by Klugman             */
   /* on page 583 at the top                              */         
   %do i=1 %to 4;
      %LimMoments&BODY_DIST(&i,_lm_tmp_xLimit, m&i);
      m&i = m&i - _lm_tmp_xLimit**&i * (1 - _lm__tmp);
   %end;

   /* This piece is so that the contribution of the body */
   /*  distribution has the correct weight,               */
   /*  i.e Param_pnLimit=empirical probability            */
   %do i=1 %to 4;
      m&i = m&i * Param_pnLimit / _lm__tmp;
   %end;


   /* To each of the moments, add the part due to the GPD tail */
   if &TAIL_PARAM_2 EQ 0 then do;
      _lm_mGPD1 = _lm_t_Param_1;
   end;
   else if 1 < 1/&TAIL_PARAM_2 then do;
      _lm_mGPD1 = (_lm_t_Param_1/&TAIL_PARAM_2) / (1/&TAIL_PARAM_2 - 1);
   end;
   else do;
      m1 = .;
   end;

   m1    = m1 + (1 - Param_pnLimit)*(_lm_mGPD1 + _lm_tmp_xLimit);
   
   if &TAIL_PARAM_2 EQ 0 then do;
      _lm_mGPD2 = 2 * _lm_t_Param_1 * _lm_t_Param_1;
   end;
   else if 2 < 1/&TAIL_PARAM_2 then do;
      _lm_mGPD2 = ((_lm_t_Param_1/&TAIL_PARAM_2)**2)*2 / ((1/&TAIL_PARAM_2 - 1) * (1/&TAIL_PARAM_2 - 2));
   end;
   else do;
      m2 = .;
   end;

   m2    = m2 + (1 - Param_pnLimit) * (_lm_mGPD2 + 2 * _lm_tmp_xLimit * _lm_mGPD1 + _lm_tmp_xLimit**2 );
   
   if &TAIL_PARAM_2 EQ 0 then do;
      _lm_mGPD3 = 6 * _lm_t_Param_1 * _lm_t_Param_1 * _lm_t_Param_1;
   end;
   else if 3 < 1/&TAIL_PARAM_2 then do;
      _lm_mGPD3 = ((_lm_t_Param_1/&TAIL_PARAM_2)**3)*6 / ((1/&TAIL_PARAM_2 - 1) * (1/&TAIL_PARAM_2 - 2) * (1/&TAIL_PARAM_2 - 3));
   end;
   else do;
      m3 = .;
   end;

   m3    = m3 + (1 - Param_pnLimit)*(_lm_mGPD3 + 3 * _lm_tmp_xLimit * _lm_mGPD2 + 3 * _lm_tmp_xLimit**2 * _lm_mGPD1  + _lm_tmp_xLimit**3 );
    
   if &TAIL_PARAM_2 EQ 0 then do;
      _lm_mGPD4 = 24 * _lm_t_Param_1 * _lm_t_Param_1 * _lm_t_Param_1 * _lm_t_Param_1;
   end;
   else if 4 < 1/&TAIL_PARAM_2 then do;
      _lm_mGPD4 = ((_lm_t_Param_1/&TAIL_PARAM_2)**4)*24 / ((1/&TAIL_PARAM_2 - 1) * (1/&TAIL_PARAM_2 - 2) * (1/&TAIL_PARAM_2 - 3) * (1/&TAIL_PARAM_2 - 4));
   end;
   else do;
      m4 = .;
   end;

   m4    = m4 + (1 - Param_pnLimit) * (_lm_mGPD4 + 4 * _lm_tmp_xLimit * _lm_mGPD3 + 6 * _lm_tmp_xLimit**2 * _lm_mGPD2 + 4 * _lm_tmp_xLimit**3 * _lm_mGPD1  + _lm_tmp_xLimit**4 );

   mean = m1;
   std  = sqrt(m2-m1*m1);

   if NOT MISSING(std) then do;
      skewness = (m3-3*m1*m2+2*(m1**3))/(std**3);
      kurtosis = (m4-4*m1*m3+6*(m1**2)*m2-3*(m1**4))/(std**4);
   end;
   else do;
      skewness=.;
      kurtosis=.;
   end;
%mend MomentsMixDist;

/*****************************************************************************************
 *   What it does  :  Does method of moments estimation and assigns the corresponding 
 *                    values to the param_1, param_2, param_4, and param_3 macros
 *
 *   No Parameters 
 *****************************************************************************************/
%macro MomFitMixDist;
    /*- expect  the mean, variance, and kurtosis of logarithm of losses -*/    
    data _null_;
        Param_2=sqrt(&variance);
        call symputX("Param_1", put(&mean, BEST20.));
        call symputX("Param_2", put(Param_2, BEST20.));
    run;
%mend MomFitMixDist;

%macro Param_BoundsMIXDIST();

%mend;
       
%macro ValidateParamsMixDist(Valid, Reason);
      %ValidateParams&body_dist(&Valid, bodyReason);

      if upcase(&Valid) = "FALSE" then do;
         Valid = &Valid;
         Reason = sasmsg("&orvdds", "dist_init_wrong_body_param_error", "quote", "_GPD_&body_dist") || " " || bodyReason;
      end;
      else do;
            if (&TAIL_PARAM_1 LE 0) then do;
               Valid  = "false";
               Reason = sasmsg("&orvdds", "dist_init_first_tail_param_neg_error", "quote", "_GPD_&body_dist");
            end;
            else if (&TAIL_PARAM_2 < 0) then do;
               valid  = "false";
               Reason = sasmsg("&orvdds", "dist_init_second_tail_param_neg_error", "quote", "_GPD_&body_dist") ;
            end;
            else if (Param_xLimit LE 0) then do;
               valid  = "false";
               Reason = sasmsg("&orvdds", "dist_init_xlimit_param_neg_error", "quote", "_GPD_&body_dist");
            end;
            else if  (Param_pnLimit LE 0) OR (Param_pnLimit GE 1)  then do;
               valid  = "false";
               Reason = sasmsg("&orvdds", "dist_init_pnlimit_outside_0_1_error", "quote", "_GPD_&body_dist");
            end;
            else do;
               Valid  = "true";
            end;
      end;
%mend ValidateParamsMixDist;
   
%macro ParamNamesMixdist(nameList=);
   %let Param_count=0;
   %let BaseName=%str();
   %orvd_&body_dist;

   %ParamNames&body_dist.(nameList=BaseName);

   data _NULL_;
      length ParamName_1-ParamName_&Param_Count $20. allNames $200;
      array bodyNames{&Body_Param_Count} &BaseName;
      array ParamNames{&Param_Count} $ ParamName_1-ParamName_&Param_Count;

      do i=1 to &Body_Param_Count;
         ParamNames(i)=vname(BodyNames(i))||"Body";
         allNames = trim(allNames) || " " || ParamNames(i);
      end;

      ParamNames(&Body_Param_Count+1) = "ScaleTail";
      ParamNames(&Param_Count)   = "ShapeTail";

      allNames=trim(allNames) || " " || "ScaleTail ShapeTail TailStart TailProb";
      call symput("&nameList",allNames);
      keep AllNames ParamName:;
   run;
%mend;

/*****************************************************************************************
 *   Needed to be called before anything else is available
 *****************************************************************************************/
%macro ORVD_MixDist(minimalInit = true);     
   %let tail_Dist            = GPD;
   %let TAIL_PARAM_COUNT     = 2;
   %let TAIL_NAME            = GPD;
   %let TAIL_SCALE_TRANSFORM = ID;

   %orvd_&body_Dist(minimalInit = &minimalInit);
   %let body_Dist = %trim(&body_Dist);

   proc sql noprint; 
      drop table &body_Dist._status; 
   quit;
   
   %let BODY_PARAM_COUNT     = &PARAM_COUNT;
   %let BODY_NAME            = &DIST_NAME;
   %let BODY_SCALE_TRANSFORM = &SCALE_TRANSFORM;
   %let BODY_DIST            = &body_Dist;

   %let TAIL_PARAM_1 = Param_%eval(&BODY_PARAM_COUNT + 1);
   %let TAIL_PARAM_2 = Param_%eval(&BODY_PARAM_COUNT + 2);

   %let PARAM_COUNT = %eval(&BODY_PARAM_COUNT + &TAIL_PARAM_COUNT );
   %let DIST_NAME   = Mixture(&BODY_NAME, &TAIL_NAME);

   %let NEEDED_INPUT    = TailBoundaryChoice TailBoundaryValue TailBoundaryPercent;
   %let SCALE_TRANSFORM = &BODY_SCALE_TRANSFORM;

   %if %upcase(&minimalInit) NE TRUE %then %do;
      proc fcmp 
               outlib = &ORVAR_CMPLIB.._GPD_&body_Dist 
               inlib  = &ORVAR_CMPLIB.;
         function Cdf__GPD_&body_Dist(x,Param_1 &&_GPD_&body_Dist._Param_List_Minus_1);
            Param_pLimit = Cdf_&body_Dist(Param_xLimit, Param_1 &&&BODY_DIST._Param_List_Minus_1);

            if x LE Param_xLimit then do;
               return(  Cdf_&body_Dist(x,Param_1 &&&BODY_DIST._Param_List_Minus_1)* Param_pnLimit/Param_pLimit);
            end;
            else do;
               %if &SCALE_TRANSFORM = ID %then %do;
                  if &TAIL_PARAM_2 LE 0.00000000001 then do;
                     return( Param_pnLimit + (1-Param_pnLimit)*(1-exp(-(x-Param_xLimit)/&TAIL_PARAM_1/Param_1)));
                  end;
                  else do;
                     return( Param_pnLimit + (1-Param_pnLimit)*(1-(1 + &TAIL_PARAM_2*(x-Param_xLimit)/(&TAIL_PARAM_1*Param_1))**(-1/&TAIL_PARAM_2)));
                  end;
               %end;
               %else %do;
                  if &TAIL_PARAM_2 LE 0.00000000001 then do;
                     return( Param_pnLimit + (1-Param_pnLimit)*(1-exp(-(x-Param_xLimit)/&TAIL_PARAM_1/exp(Param_1))));
                  end;
                  else do;
                     return( Param_pnLimit + (1-Param_pnLimit)*(1-(1 + &TAIL_PARAM_2*(x-Param_xLimit)/&TAIL_PARAM_1/exp(Param_1))**(-1/&TAIL_PARAM_2)));
                  end;
               %end;
            end;
         endsub;

         function Pdf__GPD_&body_Dist(x,Param_1 &&_GPD_&body_Dist._Param_List_Minus_1);
            Param_pLimit = Cdf_&body_Dist(Param_xLimit, Param_1 &&&BODY_DIST._Param_List_Minus_1);

            if x LE Param_xLimit then do;
               return( Pdf_&Body_Dist(x,Param_1 &&&BODY_DIST._Param_List_Minus_1)* Param_pnLimit / Param_pLimit);    
            end;
            else do;
               %if &SCALE_TRANSFORM = ID %then %do;
                  if &TAIL_PARAM_2 LE 0.00000000001 then do;
                     return((1-Param_pnLimit) * exp(-(x - Param_xLimit)/&TAIL_PARAM_1/Param_1)/&TAIL_PARAM_1/Param_1);
                  end;
                  else do;
                     return((1-Param_pnLimit)*((1 + (&TAIL_PARAM_2)*(x-Param_xLimit)/&TAIL_PARAM_1/Param_1)**(-1/&TAIL_PARAM_2 -1))/&TAIL_PARAM_1/Param_1);
                  end;
               %end;
               %else %do;
                  if &TAIL_PARAM_2 LE 0.00000000001 then do;
                     return((1-Param_pnLimit) * exp(-(x - Param_xLimit)/&TAIL_PARAM_1/exp(Param_1))/&TAIL_PARAM_1/exp(Param_1));
                  end;
                  else do;
                     return((1-Param_pnLimit)*((1 + (&TAIL_PARAM_2)*(x-Param_xLimit)/&TAIL_PARAM_1/exp(Param_1))**(-1/&TAIL_PARAM_2 -1))/&TAIL_PARAM_1/exp(Param_1));
                  end;
               %end;
            end;
         endsub;
      run;
   %end;

%DONE:
   data MixDist_STATUS;
      status = 0;
   run;
%mend ORVD_MixDist;
