/*****************************************************************************************
 * Copyright (C), 2007
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    orvinc_fittingmod                                             
 * SUPPORT: stchri
 * PURPOSE: Auto-fits a single cell for incremental load.  May refit numerous
 *          iterations for requested cell.
 *          Called by %orvinc_fitting() or %orvp_distribute_incfitting().
 *      
 * PARAMETERS:
 *          nth - used to create unique output dataset names for each host.
 *          ccuid - cell id to perform incremental fit for.
 *          See %orvinc_fitting() for rest of parameters definitions
 *
 * NOTES:   All output datasets are created with _&nth suffxi and are emptied before
 *          filling in! amcro %append_incdata at bottom fo this file can be used
 *          to appropriately append output datasets to final ones.
 *****************************************************************************************/
%macro orvinc_fittingmod(
                         nth             = ,
                         ccuid           = ,
                         ccData          = ,
                         allibls         = ,
                         CCV_NAMES       = ,
                         sparmOld        = ,
                         sparmCovOld     = ,
                         fparmOld        = ,
                         sparmNew        = ,
                         sCovNew         = ,
                         fparmNew        = ,
                         fBlEstNew       = ,
                         fCovNew         = ,
                         fModelsNew      = ,
                         lossDataNew     = ,
                         dateInc         = ,
                         kribl           = ,
                         krifirm         = ,
                         internalScaling = ,
                         globalRange     = ,
                         datatype        = ,
                         interval        = ,
                         countData       = ,
                         instrataprops   = ,
                         range_change    = true
                         );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #orvinc_fittingmod(;
      %put .                   nth             = &nth.,;
      %put .                   ccuid           = &ccuid.,;
      %put .                   ccData          = &ccData.,;
      %put .                   allibls         = &allibls.,;
      %put .                   CCV_NAMES       = &CCV_NAMES.,;
      %put .                   sparmOld        = &sparmOld.,;
      %put .                   sparmCovOld     = &sparmCovOld.,;
      %put .                   fparmOld        = &fparmOld.,;
      %put .                   sparmNew        = &sparmNew.,;
      %put .                   sCovNew         = &sCovNew.,;
      %put .                   fparmNew        = &fparmNew.,;
      %put .                   fBlEstNew       = &fBlEstNew.,;
      %put .                   fCovNew         = &fCovNew.,;
      %put .                   fModelsNew      = &fModelsNew.,;
      %put .                   lossDataNew     = &lossDataNew.,;
      %put .                   dateInc         = &dateInc.,;
      %put .                   kribl           = &kribl.,;
      %put .                   krifirm         = &krifirm.,;
      %put .                   internalScaling = &internalScaling.,;
      %put .                   globalRange     = &globalRange.,;
      %put .                   datatype        = &datatype.,;
      %put .                   interval        = &interval.,;
      %put .                   countData       = &countData.,;
      %put .                   instrataprops   = &instrataprops.,;
      %put .                   range_change    = &range_change.;
      %put .                   );
   %end;
 
   %let tmpSparm     = _sparm_&nth.; 
   %let tmpFparm     = _fparm_&nth.;

   /***** create macros for output datasets and empty them *****/
   %let outSparmNew  = &sparmNew._&nth.; 
   data &outSparmNew; set _NULL_; run;
   %let outSCovNew   = &sCovNew._&nth.;
   data &outSCovNew; set _NULL_; run;

   %if %upcase(&datatype) NE P %then %do;
   %let outFparmNew  = &fparmNew._&nth.;
   data &outFparmNew; set _NULL_; run;
   %let outFBlEstNew = &fBlEstNew._&nth.;
   data &outFBlEstNew; set _NULL_; run;
   %let outFCovNew   = &fCovNew._&nth.;
   data &outFCovNew; set _NULL_; run;
   %let outModelsNew   = &fModelsNew._&nth.;
   data &outModelsNew; set _NULL_; run;
   %end;
          
    /***** initialize user cancel status *****/
    %let user_cancel= false;
    
    /**************************************************************************************/
    /********************** loop over SEVERITY iterations for our cell ********************/
    /**************************************************************************************/
   
    /***** subset sparm dataset to get only the iterations for our cell 
           and number them so we can loop *****/
    %let numIters = 0;
    data &tmpSparm; set &sparmOld end= LAST; 
      where CCUID eq &ccuid; 
      N = _N_;
      if LAST then do;
        call symput("numIters", _N_);
      end;
    run;
 
    /***** loop over all iterations in our cell *****/
    %let sev_was_fit = &orv_false.;

    %do parm_i = 1 %to &numIters;
         %let append_seviter = false;

         /***** get current Severity Iteration *****/
         %let sevIter= 0;
         %let status= 1;
         data _sparm_init; set &tmpSparm; where N eq &parm_i;
            call symputX('sevIter', trim(left(iteration)));
            if selection eq 'S' then
                  call symputX('sevSelected', 1);
            else  call symputX('sevSelected', 0);
            call symput('estType', ESTTYPE);
            call symput('cellChanged', change);
            call symputX('status', status);
            drop N CHANGE;
         run;
         
         %if &status ne 1 %then %do;
            /***** No re-fit needed (just copy over) *****/
            /***** Scenarios have estType 'M' (moments estimation) or 'P' (parm estimation) *****/
            %if &estType eq M or &estType eq P or &cellChanged ne 1 %then %do;      
               %if &estType eq M or &estType eq P %then %do;
                   %let scenarioOnly = true;
                   %put &orvar_status %sysfunc(sasmsg(&orvdds, fmod_copying_sev_scen_status, noquote, &ccUid, &sevIter));
               %end;
               %else %do;
                   %let scenarioOnly = false;
                   %put &orvar_status %sysfunc(sasmsg(&orvdds, fmod_skipping_sev_afit_status, noquote, &ccUid, &sevIter));
               %end;
               
               /* get this iterations covariance matrix */
               %let nobs=0;
               %orvu_getObsCount(dataIn=&sparmCovOld,theObsCount=nobs);
               %if &nobs NE 0 %then %do;
               data _tmpscov; set &sparmCovOld; 
                  where CCUID = &ccuid and ITERATION = &sevIter; 
               run;
               %end;
               %else %do;
                  data _tmpscov; set _NULL_; run; 
               %end;
               
               %orva_append_parm_dataset(
                     tmpParm           =_sparm_init,
                     finalParm         =&outSparmNew,
                     tmpCov            =_tmpscov,
                     finalCov          =&outSCovNew,
                     ccUid             =&ccUid, 
                     iteration         =%eval(&sevIter-1),
                     sevIteration      =,
                     sevIterSelected   =&sevSelected,
                     measCol           =LOGL 1 KS_V 0 SBC 0 AIC 0 CVM_V 0 AD_V 0 ,
                     selectedMeas      =LOGL,
                     scenarioOnly      =&scenarioOnly,
                     rank              =false);
            %end;
            /***** Need to re-fit *****/      
            %else %do;
               %put &orvar_status %sysfunc(sasmsg(&orvdds, fmod_performing_sev_afit_stauts, noquote, &ccUid, &sevIter));              
               %orva_sfit(  data             =&lossDataNew,   
                            datatype         =&datatype,    
                            ccUid            =&ccUid,
                            ccData           =&ccData,
                            initData         =_sparm_init,
                            outparm          =&outSparmNew,
                            outcov           =&outSCovNew,
                            iteration        =&sevIter
                          );                      
            %end;
            data _NULL_;
               set &outSparmNew; 
               where CCUID = &ccUid and ITERATION eq &sevIter;
               if Status eq 0 then do; 
                  call symputX("sev_was_fit", "&orv_true.");
               end;
            run; 
            
         %end; /***** previous fit is valid *****/
   
         /***** check for user requested cancel *****/
         %orvp_user_cancel();
         %if &user_cancel eq true %then %do; %goto DONE; %end;
   
    %end; /***** looping over SEVERITY iterations for our cell *****/
    
   /**************************************************************************************/
   /********************* loop over FREQUENCY iterations for our cell ********************/
   /**************************************************************************************/
   /***** subset fparm dataset to get only the iterations for our cell 
           and number them so we can loop *****/
   %if %upcase(&datatype) NE P %then %do;   
      %let numIters = 0;

      data &tmpFparm; set &fparmOld end= LAST; 
         where CCUID eq &ccuid; 
         N = _N_;
         if LAST then do;
           call symput("numIters", _N_);
         end;
      run;
 
      /***** loop over all iteration in our cell *****/
      %do parm_i = 1 %to &numIters;
         %let append_freqiter = false;    
         
         /***** get current Frequency Iteration *****/
         %let countKriFreq= 0;
         %let freqIter= 0;      
         data _fparm_init; set &tmpFparm; 
              where N eq &parm_i;
              call symputX('sevIter', SEVITER);
              call symputX('countKriFreq', count_kri);
              call symput('freqKris', selectedKris);
              call symputX('FreqIter', trim(left(iteration)));
              if selection eq 'S' then
                    call symputX('freqSelected', 1);
              else  call symputX('freqSelected', 0);
              call symput('fdistribution', distribution);
         run;  
         
         /***** Is Iteration a Scenario? (nothing to do) *****/
         /***** Scenarios have estType 'M' (moments estimation) or 'P' (parm estimation) *****/
         /***** also check for no cell changes AND no range change *****/
         %if &estType eq M or &estType eq P or
             ( &range_change ne true AND &cellChanged ne 1) %then %do;
             
             %if &estType eq M or &estType eq P %then %do;
               %let scenarioOnly = true;
               %put &orvar_status %sysfunc(sasmsg(&orvdds, fmod_freq_scen_copied_status, noquote, &ccUid, &freqIter));
             %end;
             %else %do;
                %let scenarioOnly = false;
                %put &orvar_status %sysfunc(sasmsg(&orvdds, fmod_skipping_freq_afit_status, noquote, &ccUid, &freqIter));
             %end;

             %orva_append_parm_dataset(
                     tmpParm           =_fparm_init,
                     finalParm         =&outFparmNew,
                     ccUid             =&ccUid, 
                     iteration         =%eval(&freqIter-1),
                     sevIteration      =%eval(&sevIter),
                     sevIterSelected   =&sevSelected,
                     measCol           =LOGL 1 SBC 0 AIC 0 ,
                     selectedMeas      =LOGL,                  
                     scenarioOnly      =&scenarioOnly,
                     rank              =false);

         %end;
         
         /***** Need to re-fit *****/      
         %else %do;
            /*- get associated sparm iteration -*/
            %let sevIterSelected=0;
             %let sevStatus = 0;
            data _sparm_init; set &outSparmNew; where CCUID=&ccUid and ITERATION eq &sevIter; 
                 call symputx("sevStatus", status); /*should have only one record*/
              if SELECTION eq 'S' then do;
                call symputx("sevIterSelected", 1);
              end;
            run; 
            %let nobs=0;
            %let dsid=%sysfunc(open(_sparm_init,I));
            %if &dsid ^= 0 %then %do;
               %let nobs=%sysfunc(attrn(&dsid,nlobs));
               %let rc=%sysfunc(close(&dsid));
            %end;
          %if (&nobs NE 0) AND (&sevStatus NE 1) %then %do;
               %put &orvar_status %sysfunc(sasmsg(&orvdds, fmod_freq_fit_status, noquote, &ccUid, &freqIter));

               %orva_freq_fit( 
                              countdata        = &countData,
                              lossdata         = &lossDataNew,
                              initData         = _fparm_init, 
                              iteration        = &freqIter,
                              instrataprops    = &instrataprops,
                              useCellOverrides = false,  
                              cTool            = false,
                              interval         = &interval,
                              pobservInfo      = _sparm_init,
                              seviteration     = &sevIter,                          
                              sevIterSelected  = &sevIterSelected,
                              datatype         = &datatype, 
                              ccUid            = &ccUid, 
                              ccData           = &ccData,   
                              allibls          = &allibls.,
                              CCV_NAMES        = &CCV_NAMES,
                              outParm          = &outFParmNew,
                              outBluidEst      = &outFBlEstNew,
                              outCov           = &outFCovNew,
                              outModels        = &outModelsNew                        
                              );            
            %end;
            %else %do;
               %put &orvar_status %sysfunc(sasmsg(&orvdds, fmod_sev_fit_doesnt_exist_error, noquote, &ccUid, &parm_i));
               %let append_freqiter = false;
            %end;
         %end;
 
         /***** check for user requested cancel *****/
         %orvp_user_cancel();

         %if &user_cancel eq true %then %do; 
            %goto DONE; 
         %end;
   
      %end; /***** looping over FREQUENCY iterations for our cell *****/

      %if &sev_was_fit. eq &orv_true. %then %do;
         /* Make sure we fit a Frequency distribution to the selected severity iteration ------------*/
         %_ovr_check_selected_severity( 
                                        countdata        = &countData,
                                        lossdata         = &lossDataNew,                              
                                        instrataprops    = &instrataprops,
                                        interval         = &interval,
                                        datatype         = &datatype, 
                                        ccUid            = &ccUid, 
                                        ccData           = &ccData,   
                                        allibls          = &allibls.,
                                        CCV_NAMES        = &CCV_NAMES,
                                        outFParm         = &outFParmNew,
                                        outSParm         = &outSParmNew,
                                        outBluidEst      = &outFBlEstNew,
                                        outCov           = &outFCovNew,
                                        outModels        = &outModelsNew 
                                        );
      %end;
   %end; /* datatype ne P */ 
%DONE:
       
   /***** cleanup temporary datasets *****/
   %orvu_delete_data( data =
                            _sparm_init
                            &tmpSparm     
                            %if %upcase(&datatype) NE P %then %do;
                               _fparm_init
                               tmpf_blest_&nth
                               tmpf_cov_&nth
                               &tmpFparm
                            %end;
                            );
%mend orvinc_fittingmod;
   
%macro getFreqKriVars(freqKris = );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @getFreqKriVars(freqKris = &freqKris.);
   %end;

    %local countKriFreq kri last2;
    data _NULL_;
       length freqKrisFirm $ 1024;
       length freqKrisBl $ 1024;
       nKrisFirm = 0;
       nKrisBl = 0;
       freqKrisFirm = "";
       freqKrisBl = "";
       %let countKriFreq=1;
       %let kri=%scan(&freqKris, &countKriFreq, ' ');
       %do %while (&kri ne );
           %let last2 = %substr(&kri,%length(&kri)-1);
           %if "&LAST2" EQ "_F"  %then %do;
                freqKrisFirm = trim(left(freqKrisFirm)) || ' ' || trim(left("&kri"));
                nKrisFirm = nKrisFirm + 1;
           %end;
           %else %do;
                freqKrisBl = trim(left(freqKrisBl)) || ' ' || trim(left("&kri"));
                nKrisBl = nKrisBl + 1;
           %end;
           %let countKriFreq=%eval(&countKriFreq+1);
           %let kri=%scan(&freqKris, &countKriFreq, ' ');
       %end;
       call symput('freqKrisFirm', freqKrisFirm);
       call symput('freqKrisBl', freqKrisBl);
       call symput('nKrisFirm', nKrisFirm);
       call symput('nKrisBl', nKrisBl);
    run;
%mend getFreqKriVars;

/*
 * Since this macro always create new output datasets, the caller must append 
 * outputs to their final datasets.
 */
%macro orvinc_append_incdata(
                             incData   = , 
                             finalData = , 
                             nth       =
                             );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put @orvinc_append_incdata(;
      %put .                       incData   = &incData.,; 
      %put .                       finalData = &finalData.,; 
      %put .                       nth       = &nth.;
      %put .                       );
   %end;

   %let nobs = 0;
   %orvu_getObsCount(&incData._&nth., nobs);
   %if &nobs NE 0 %then %do;
      %let nobs = 0;
      %orvu_getObsCount(&finalData, nobs);
      %if &nobs EQ 0 %then %do;           
          data &finalData(compress = binary); set &incData._&nth.; run;
      %end;
      %else %do;
          data &finalData(compress = binary); set &finalData &incData._&nth.; run;
      %end;
   %end;
%mend orvinc_append_incdata;


/* Make sure we fit a Frequency distribution to the selected severity iteration ------------*/
%macro _ovr_check_selected_severity( 
                                    countdata     = ,
                                    lossdata      = ,                              
                                    instrataprops = ,
                                    interval      = ,
                                    datatype      = , 
                                    ccUid         = , 
                                    ccData        = ,   
                                    allibls       = ,
                                    CCV_NAMES     = ,
                                    outFParm      = ,
                                    outSParm      = ,
                                    outBluidEst   = ,
                                    outCov        = ,
                                    outModels     = 
                                    );

   /* Find selected severity -------------------------------------*/
   %let sevIter = -1;
   data _sparm_init; 
      set &outSparm; 
      where CCUID = &ccUid and SELECTION eq 'S';
      call symputx("sevIter", ITERATION);
      call symputx("sevStatus", status); /*should have only one record*/
   run; 

   %if &sevIter = -1 %then %do;
      /* No severity iterations selected -----*/
      %return;
   %end;

   /* Check if it has a frequency model selected -----------------*/
   %local found;
   %let found = false;
   %if %orvu_var_exists( ccuid, &outFParm) ne 0 %then %do;
      data _NULL_; 
         set &outFParm;
         where CCUID = &ccUid and seviter = &sevIter and SELECTION eq 'S';
         call symput("found", "true");
      run;
   %end;

   /* Fit one if none selected -----------------------------------*/
   %if &found = false %then %do;

      /* fitopts will always be in varprj_m as the code is written now, 
      because it is placed there at project creation (copied from the 
      base project project). To be on the safe side, check in vaprj
      folder in case we change the way fitopts creation works - can't hurt. */
      %local fitoptsDataset;
      %orvu_find_dataset(
                         nameIn  = fitopts, 
                         Libs    = VARPRJ VARPRJ_M, 
                         nameOut = fitoptsDataset
                         );
      %orva_freq_fit(          
                     countdata        = &countData,
                     lossdata         = &lossDataNew,
                     fitoptsdata      = &fitoptsDataset,
                     initData         =, 
                     iteration        =,
                     instrataprops    = &instrataprops,
                     useCellOverrides = true,  
                     cTool            = false,
                     interval         = &interval,
                     pobservInfo      = _sparm_init,
                     seviteration     = &sevIter,                          
                     sevIterSelected  = 1,
                     datatype         = &datatype, 
                     ccUid            = &ccUid, 
                     ccData           = &ccData,   
                     allibls          = &allibls.,
                     CCV_NAMES        = &CCV_NAMES,
                     outParm          = &outFParm,
                     outBluidEst      = &outBluidEst,
                     outCov           = &outCov,
                     outModels        = &outModels                        
                     );     
   %end;
 
%mend _ovr_check_selected_severity;
