/*************************************************************************************
 * Copyright (C), 2007
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME   : orvp_distribute_incfitting                                             
 * PURPOSE: Macro to distribute the %orva_incfitting() macro on different 
 *          machines (or cpus) by cell.  
 *                    
 * SUPPORT: stchri
 *
 * PARAMETERS:
 *      See documentation for %orva_incfitting.
 * 
 * NOTES:  This macros should only be called by %orva_incfitting and assumes
 *         someone has already called %orvp_signon().
 *
 **************************************************************************************/
%macro orvp_distribute_incfitting( 
                                   cells           =,
                                   ncells          =,
                                   sparmOld        =,
                                   sparmCovOld     =,
                                   allibls         =,
                                   fparmOld        =,
                                   sparmNew        =,
                                   sCovNew         =,
                                   fparmNew        =,
                                   fBlEstNew       =,
                                   fCovNew         =,
                                   fModelsNew      =,
                                   lossDataNew     =,
                                   ccData          =,
                                   dateInc         =,
                                   kribl           =,
                                   krifirm         =,
                                   internalScaling =,
                                   globalRange     =,
                                   datatype        =,
                                   interval        =,
                                   fitstatus       =,
                                   countData       =,
                                   instrataprops   =,
                                   CCV_NAMES       =
                                   );

   /*--- preset STATUS and REASON macro vars ---*/
   %global ORVP_STATUS;
   %let ORVP_STATUS = 1;

   %global ORVP_REASON;
   %let ORVP_REASON = "%sysfunc(sasmsg(&orvlds, unknown_error, noquote))";

   /*--- save sysem options ---*/
   %local notesopt; 
   %let notesopt = %sysfunc(getoption(notes));

   %local mprintopt; 
   %let mprintopt = %sysfunc(getoption(mprint));

   %local symbolgenopt; 
   %let symbolgenopt = %sysfunc(getoption(symbolgen));
    
   /*--- set initialized flag to 0 for each host ---*/
   %do iHost = 1 %to &NumHosts;
      data _null_;
         inited = 0;
         call symput('Inited'||trim(left("&iHost")),inited);
      run;
   %end;

   /* -- set the libname path for the _cancel_ data set ------------------*/
   %let varprj_path   = %sysfunc(pathname(varprj));
   %let varprj_M_path = %sysfunc(pathname(varprj_M));

   /*--- loop over CCUIDs in sparm dataset (fittings) ---*/  
   %let ITER = 1;
   %let NREPSLEFT = &ncells;

   %do %while (&NREPSLEFT gt 0);
      /*--- loop over available hosts ---*/
      %let num_submitted = 0;

      %do iHost = 1 %to &NumHosts;
         %if (&&Status&iHost ^= -1) and (&NREPSLEFT gt 0) %then %do;
            /*--- get next cell (CCUID) to process ---*/
            data _NULL_; 
               set &cells; 
               where N eq &ITER; 
               call symput('ccuid', trim(left(CCUID))); 
            run;

            %orvp_cell_status(
                               statusDS = &fitstatus,
                               ccuid    = &ccuid,
                               status   = 0
                               );
 
            options remote = Host&iHost;
            
            /*--- push required macro variables to host ---*/ 
            %syslput rem_notesopt       = &notesopt;  
            %syslput rem_mprintopt      = &mprintopt;  
            %syslput rem_symbolgenopt   = &symbolgenopt;  
            %syslput rem_NTH            = &iHost;  

            %syslput  ccuid             = &ccuid;
            %syslput  CCV_NAMES         = &CCV_NAMES;
            %syslput  dateInc           = &dateInc;
            %syslput  datatype          = &datatype;
            %syslput  interval          = &interval;
            %syslput  varprj_path       = &varprj_path;
            %syslput  varprj_M_path     = &varprj_M_path;

            /*-- set up macros for uploaded datasets on hosts --*/
            %if &sparmOld NE %str() %then %do;            
               %syslput  rem_sparmOld = &sparmOld;
            %end;

            %if &sparmCovOld NE %str() %then %do;            
               %syslput  rem_sparmCovOld = &sparmCovOld;
            %end;

            %if &allibls NE %str() %then %do;            
               %syslput  rem_allibls = &allibls;
            %end;

            %if &fparmOld NE %str() %then %do;             
               %syslput  rem_fparmOld = &fparmOld;
            %end;

            %if &lossDataNew NE %str() %then %do;            
                  %syslput  rem_lossDataNew = &lossDataNew;
            %end;

            %if &ccData NE %str() %then %do;             
               %syslput  rem_ccData = &ccData;
            %end;

            %if &countData NE %str() %then %do;
               %syslput  rem_countData = &countData;
            %end;

            %if &instrataprops NE %str() %then %do;             
               %syslput  rem_instrataprops = &instrataprops;
            %end;

            %if &kribl NE %str() %then %do;             
               %syslput  rem_kribl = &kribl;
            %end;

            %if &krifirm NE %str() %then %do;           
               %syslput  rem_krifirm = &krifirm;
            %end;

            %if &internalScaling NE %str() %then %do;            
               %syslput  rem_internalScaling = &internalScaling;
            %end;

            %if &globalRange NE %str() %then %do;            
               %syslput  rem_globalRange = &globalRange;
            %end;
            
            /*--- submit code on Host ---*/
            rsubmit 
                  remote = Host&iHost 
                  cwait  = no
                  ;
               /*--- transfer system options to host ---*/
               options 
                        &rem_notesopt 
                        &rem_mprintopt 
                        &rem_symbolgenopt
                        ;

               /*--- upload needed datasets to host (just once) ---*/
               %if &&Inited&iHost eq 0 %then %do;   
                  %if &sparmOld NE %str() %then %do;
                     proc upload 
                                 data = &rem_sparmOld 
                                 out  = _remSparmOld; 
                     run;              
                  %end;             
 
                  %if &sparmCovOld NE %str() %then %do;
                     proc upload 
                                 data = &rem_sparmCovOld 
                                 out  = _remSparmCovOld; 
                     run;              
                  %end; 
 
                  %if &allibls NE %str() %then %do;  
                     proc upload 
                                 data = &rem_allibls 
                                 out  = _remallibls; 
                     run;   
                  %end;

                  %if &fparmOld NE %str() %then %do;
                     proc upload 
                                 data = &rem_fparmOld 
                                 out  = _remFparmOld; 
                     run;                               
                  %end;           

                  %if &lossDataNew NE %str() %then %do;
                     proc upload 
                                 data = &rem_lossDataNew 
                                 out  = _remLossData; 
                     run;                               
                  %end;   

                  %if &ccData NE %str() %then %do;
                     proc upload 
                                 data = &rem_ccData 
                                 out  = _remCcData;
                     run;               
                  %end;              

                  %if &countData NE %str() %then %do;
                     proc upload 
                                 data = &rem_countData 
                                 out  = _remcountData; 
                     run;               
                  %end;              

                  %if &instrataprops NE %str() %then %do;
                     proc upload 
                                 data = &rem_instrataprops 
                                 out  = _remInstrataprops; 
                     run;               
                  %end;              
             
                  %if &kribl NE %str() %then %do;
                     proc upload 
                                 data = &rem_kribl 
                                 out  = _remKribl; 
                     run;
                  %end;              

                  %if &krifirm NE %str() %then %do;
                     proc upload 
                                 data = &rem_krifirm 
                                 out  = _remKrifirm; 
                     run;                
                  %end;              

                  %if &internalScaling NE %str() %then %do;
                     proc upload 
                                 data = &rem_internalScaling 
                                 out  = _remInternalScaling; 
                     run;               
                  %end;              

                  %if &globalRange NE %str() %then %do;
                     proc upload 
                                 data = &rem_globalRange 
                                 out  = _remGlobalRange; 
                     run;              
                  %end;           
               %end;

               /* Setup varprj for _cancel_ data set --------------------*/
               libname varprj   "&varprj_path";
               libname varprj_M "&varprj_M_path";
                                                              
               /*--- submit fitting code ---*/
               %orvinc_fittingmod(
                                    nth             = &rem_NTH,
                                    ccuid           = &ccuid,
                                    allibls         = _remallibls,
                                    sparmOld        = _remSparmOld,
                                    sparmCovOld     = _remSparmCovOld,
                                    fparmOld        = _remFparmOld,
                                    sparmNew        = _tmpSparmNew,
                                    scovNew         = _tmpSCovNew,
                                    fparmNew        = _tmpFparmNew,
                                    fBlEstNew       = _tmpFBlEstNew,
                                    fCovNew         = _tmpFCovNew,
                                    fModelsNew      = _tmpFModelsNew,
                                    lossDataNew     = _remLossData,
                                    ccData          = _remCcData,
                                    dateInc         = &dateInc,
                                    kribl           = _remKribl,
                                    krifirm         = _remKrifirm,
                                    internalScaling = _remInternalScaling,
                                    globalRange     = _remGlobalRange,
                                    datatype        = &datatype,
                                    interval        = &interval,
                                    countData       = _remCountData,
                                    instrataprops   = _remInstrataprops,
                                    CCV_NAMES       = &CCV_NAMES
                                    );
                
               /*--- append output datasets on host ---*/             
               %orvinc_append_incdata(
                                        incData   = _tmpSparmNew, 
                                        nth       = &rem_NTH, 
                                        finalData = _remSparmNew_&rem_NTH
                                        );

               %orvinc_append_incdata(
                                        incData   = _tmpSCovNew, 
                                        nth       = &rem_NTH, 
                                        finalData = _remSCovNew_&rem_NTH
                                        );
               
               %if &datatype ne P %then %do;
                  %orvinc_append_incdata(
                                          incData   = _tmpFparmNew, 
                                          nth       = &rem_NTH, 
                                          finalData = _remFparmNew_&rem_NTH
                                          );
                  %orvinc_append_incdata(
                                          incData   = _tmpFBlEstNew, 
                                          nth       = &rem_NTH, 
                                          finalData = _remFBlEstNew_&rem_NTH
                                          );

                  %orvinc_append_incdata(
                                          incData   = _tmpFCovNew, 
                                          nth       = &rem_NTH, 
                                          finalData = _remFCovNew_&rem_NTH
                                          );

                  %orvinc_append_incdata(
                                          incData   = _tmpFModelsNew, 
                                          nth       = &rem_NTH, 
                                          finalData = _remFModelsNew_&rem_NTH
                                          );
               %end;
            endrsubmit;
                              
            /*--- annotate that we have already initialized this host ---*/
            data _NULL_;
               done = 1;
               call symput("Inited&iHost", done);
            run;

            %let NREPSLEFT = %eval(&NREPSLEFT - 1);
            %let ITER = %eval(&ITER + 1);
            %let num_submitted = %eval(&num_submitted + 1);
         %end; /* NREPSLEFT for host */
      %end; /* end of host loop */
   
      /*--- wait for submission on each host to complete ---*/
      %do iHost = 1 %to &NumHosts;
         %if &&Status&iHost ne -1 and &num_submitted > 0 %then %do;
            rget remote = Host&iHost;

            data _NULL_; 
               set &cells; 
               where N eq &ITER - &iHost; 
               call symput('ccuid', trim(left(CCUID))); 
            run;

            %orvp_cell_status(
                               statusDS = &fitstatus,
                               ccuid    = &ccuid,
                               status   = 1
                               );

            %let num_submitted = %eval(&num_submitted - 1);
         %end;
      %end;
   
      /*--- check for user requested cancel ---*/
      %orvp_user_cancel();

      %if &user_cancel eq true %then %do; 
         %goto DONE; 
      %end;                  
   %end; /*--- end of ccuid loop ---*/

%DONE:
   /*--- collect our resultant data ---*/
   %IncFit_Collect(_remSparmNew,   &sparmNew,   &NumHosts);
   %IncFit_Collect(_remSCovNew,    &sCovNew,    &NumHosts);
   %IncFit_Collect(_remFparmNew,   &fparmNew,   &NumHosts);
   %IncFit_Collect(_remFBlEstNew,  &fBlEstNew,  &NumHosts);
   %IncFit_Collect(_remFCovNew,    &fCovNew,    &NumHosts);
   %IncFit_Collect(_remFModelsNew, &fModelsNew, &NumHosts);
  
   %let ORVP_STATUS = 0;
   %let ORVP_REASON = "%sysfunc(sasmsg(&orvlds, success_note, noquote))";
%mend orvp_distribute_incfitting;
  
%macro IncFit_Collect(from, to, n); 
   %let tonobs = 0;
   %orvu_getObsCount(&to, tonobs);

   data &to; 
      set 
         %if &tonobs NE 0 %then %do; 
            &to 
         %end;

         %let iteration = 1;

         %do %while (&iteration le &n); 
            %do iHost = 1 %to &NumHosts;
               %if &&Status&iHost ne -1 and &iteration le &n %then %do;
                  RWork&iHost..&from._&iteration    
                  %let iteration = %eval(&iteration + 1);
               %end;
            %end;        
         %end;
         ;
   run;

   %do iHost = 1 %to &NumHosts;
      %if &&Status&iHost ne -1 %then %do;
         %orvu_delete_data( 
                           data = &from.:,
                           lib  = RWork&iHost.
                           );
      %end;
   %end;        
%mend IncFit_Collect;
