/*************************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME   : orvp_distribute_loss_sim                                             
 * PURPOSE: Macro to distribute the orva_loss_sim_stream() macro on different machines.
 *          Should only be called by orva_loss_sim()!             
 * SUPPORT: stchri
 *
 * PARAMETERS :
 * 
 **************************************************************************************/
%macro orvp_distribute_loss_sim(
                                nreps       = ,
                                nstreams    = ,
                                seedData    = ,
                                isTotalCell = 
                                );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;
      %put #orvp_distribute_loss_sim(;
      %put .                          nreps       = &nreps.,;
      %put .                          nstreams    = &nstreams.,;
      %put .                          seedData    = &seedData.;
      %put .                          isTotalCell = &isTotalCell.;
      %put .                          );
   %end;

   /*--- preset STATUS and REASON macro vars ---*/
   %global ORVP_STATUS;
   %let ORVP_STATUS = 1;

   %global ORVP_REASON;
   %let ORVP_REASON = "%sysfunc(sasmsg(&orvlds, unknown_error, noquote))";

   %local notesopt; 
   %let notesopt = %sysfunc(getoption(notes));

   %local mprintopt; 
   %let mprintopt = %sysfunc(getoption(mprint));

   %local symbolgenopt; 
   %let symbolgenopt = %sysfunc(getoption(symbolgen));

   libname onLocal "%sysfunc(pathname(varprj))";
   libname finalD "%sysfunc(pathname(varprj))";    

   data onLocal._loss_;
      set _loss_;
   run;
   %if "&AVERAGE_KRI_EFFECT" NE "1" %then %do;
      data onLocal._tmp_kri_;
         set _tmp_kri_;
      run;
   %end;
 
   /*--- set initialized flag to 0 for each host ---*/
   %do iHost = 1 %to &NumHosts;
      data _null_;
         inited = 0;
         call symput('Inited'||trim(left("&iHost")),inited);
      run;
   %end;
   
   /*--- get SEED for each stream ---*/
   %do i = 1 %to &nstreams
      %let SEED = 0;

      data _NULL_; 
         set &seedData; 
         where CCUID eq &ccUID and HOST eq &i;
         length tseed $ 32;
         tseed = put(seed,32.);
         call symput('SEED'||trim(left("&i")), tseed);
      run;    
   %end;

   /*--- loop over Streams ---*/ 
   data _NULL_;
      reps = ceil(&nreps/&nstreams);
      call symput('REPS', reps);
   run;

   %let AllReps   = &reps;
   %let CumReps   = &reps;
   %let NITER     = 1;
   %let STARTREP  = 1;
   %let NREPSLEFT = &nreps;

   %do %while (&NREPSLEFT gt 0);
      /*--- loop over available hosts ---*/
      %let num_submitted = 0;

      %do iHost = 1 %to &NumHosts;
         %if (&&Status&iHost ^= -1) and (&NREPSLEFT gt 0) %then %do;
            options remote = Host&iHost;

            /*--- push required macro variables to host ---*/ 
            %syslput rem_notesopt        = &notesopt;  
            %syslput rem_mprintopt       = &mprintopt;  
            %syslput rem_symbolgenopt    = &symbolgenopt;  

            %syslput rem_SEED            = &&SEED&NITER;  
            %syslput rem_STARTREP        = &STARTREP;  
            %syslput rem_REPS            = &REPS;  
            %syslput rem_NITER           = &NITER;  

            /*----   Moving results from orva_insurance_get_effect    ----*/
            %syslput CCUid                   = &CCUid.; 
            %syslput SimDate                 = &simDate.;
            %syslput pInsurance              = &pInsurance.;
            %syslput pInsuranceCount         = &pInsuranceCount.;
            %syslput Insurance_UID_1         = &Insurance_UID_1.;
            %syslput Insurance_Index_1       = &Insurance_Index_1.;
            %syslput Deductible_1            = &Deductible_1.;
            %syslput The_Limit_1             = &The_Limit_1.;
            %syslput Eff_Agg_Limit_1         = &Eff_Agg_Limit_1.;
            %syslput Payout_Haircut_1        = &Payout_Haircut_1.;
            %syslput PolicyExpiryDate_1      = &PolicyExpiryDate_1.;
            %syslput Insurance_UID_2         = &Insurance_UID_2.;
            %syslput Insurance_Index_2       = &Insurance_Index_2.;
            %syslput Deductible_2            = &Deductible_2.;
            %syslput The_Limit_2             = &The_Limit_2.;
            %syslput Eff_Agg_Limit_2         = &Eff_Agg_Limit_2.;
            %syslput Payout_Haircut_2        = &Payout_Haircut_2.;
            %syslput PolicyExpiryDate_2      = &PolicyExpiryDate_2.;
            %syslput coverageData1           = &coverageData1.;
            %syslput CoverageCount           = &CoverageCount.;
            %syslput InsuranceContractCount  = &InsuranceContractCount.;
            %syslput isTotalCell             = &isTotalCell.;
            /*-----------------------------------------------------------*/

            %syslput lookupSize          = &lookupSize;
            %syslput LOOKUP_QUANTILES    = &LOOKUP_QUANTILES;
            %syslput MAXLOSS             = &MAXLOSS;
            %syslput AVERAGE_KRI_EFFECT  = &AVERAGE_KRI_EFFECT;

            %if "&AVERAGE_KRI_EFFECT" ne "1" %then %do;
               %syslput KRI_CHOICES      = &KRI_CHOICES;
            %end;

            %syslput PInsurance          = &PInsurance;
            %syslput InsuranceArray      = &InsuranceArray;
            %syslput simDate             = &simDate;
            %syslput remgpath            = %sysfunc(compress(%sysfunc(pathname(onlocal)),'()'));

            %if "&coverageData1" ne "" and &CoverageCount ^= 0 %then %do;
               %syslput pInsuranceCount        = &pInsuranceCount;
               %syslput InsuranceContractCount = &InsuranceContractCount;
               %syslput Insurance_UID_1        = &Insurance_UID_1;      
               %syslput Insurance_Index_1      = &Insurance_Index_1;      
               %syslput Deductible_1           = &Deductible_1;      
               %syslput The_Limit_1            = &The_Limit_1;                      
               %syslput Eff_Agg_Limit_1        = &Eff_Agg_Limit_1;     
               %syslput Payout_Haircut_1       = &Payout_Haircut_1;    
               %syslput Insurance_UID_2        = &Insurance_UID_2;      
               %syslput Insurance_Index_2      = &Insurance_Index_2;      
               %syslput Deductible_2           = &Deductible_2 ;            
               %syslput The_Limit_2            = &The_Limit_2;                      
               %syslput Eff_Agg_Limit_2        = &Eff_Agg_Limit_2;              
               %syslput Payout_Haircut_2       = &Payout_Haircut_2; 
               %syslput PolicyExpiryDate_1     = &PolicyExpiryDate_1;         
               %syslput PolicyExpiryDate_2     = &PolicyExpiryDate_2; 
            %end;

            %syslput insuranceMethod     = &insuranceMethod; 
           
            /* needed for rare event simulation */
            %syslput TOTAL_SCEN_MEAN     = &TOTAL_SCEN_MEAN ;

            /*--- submit fit code on Host ---*/
            rsubmit remote = Host&iHost cwait = no;             
               /*--- transfer system options to host ---*/
               options 
                       &rem_notesopt 
                       &rem_mprintopt 
                       &rem_symbolgenopt
                       ;

               
               libname finalD "&remGPath";

               /*--- upload needed data to host ---*/
               %if &&Inited&iHost eq 0 %then %do;  
                  %if "&AVERAGE_KRI_EFFECT" NE "1" %then %do;
                     data work._tmp_kri_; 
                        set finalD._tmp_kri_;
                     run; /* proc upload data=_tmp_kri_; run; */
                  %end;

                  data work._loss_;
                     set finalD._loss_;
                  run;    /*  proc upload data=_loss_; run; */               
               %end;

               %orva_loss_sim_stream( 
                                     startRep        = &rem_STARTREP,
                                     nreps           = &rem_REPS,
                                     streamNum       = &rem_NITER,
                                     seed            = &rem_SEED, 
                                     isTotalCell     = &isTotalCell,
                                     insuranceMethod = &insuranceMethod,
                                     finald          = finald
                                     );
            endrsubmit;

            data _NULL_;
               done = 1;
               call symput("Inited&iHost", done);
            run;

            %let STARTREP  = %eval(&STARTREP + &REPS);
            %let NREPSLEFT = %eval(&NREPSLEFT - &REPS);

            %if &NREPSLEFT lt &REPS %then %do;
               %let REPS = &NREPSLEFT;
            %end;

            %let NITER = %eval(&NITER + 1);
            %let num_submitted = %eval(&num_submitted + 1);
            %let CumReps = %eval(&CumReps + &reps);
            %let AllReps = &AllReps &CumReps;
         %end;
      %end; /* end of host loop */
   
      /*--- wait for submission in each host to complete ---*/
      %do iHost = 1 %to &NumHosts;
         %if &&Status&iHost ne - 1 and &num_submitted > 0 %then %do;
            rget remote = Host&iHost;
            %let num_submitted = %eval(&num_submitted - 1);
         %end;
      %end;
   %end; /*--- end of distribution loop ---*/

   /*--- collect our result data ---*/
   %LossSim_Collect(
                    _osimNaked_, 
                    _osimNaked_, 
                    &nstreams
                    );
   %LossSim_Collect(
                    _osimInsured_, 
                    _osimInsured_, 
                    &nstreams
                    );

   %orvu_delete_data( 
                     data = _loss_,
                     lib  = finalD
                     );

   %let ORVP_STATUS = 0;
   %let ORVP_REASON = "%sysfunc(sasmsg(&orvlds, success_note, noquote))";
%mend orvp_distribute_loss_sim;

%macro LossSim_Collect(from, to, n);
   data &to; 
      set 
         %let iteration = 1;

         %do %while (&iteration le &n); 
            %do iHost = 1 %to &NumHosts;
               %if &&Status&iHost ne -1 and &iteration le &n %then %do;
                  finalD.&from.&iteration   
                  %let iteration = %eval(&iteration + 1);
               %end;
            %end;        
         %end;
         ;
   run;

   %orvu_delete_data( 
                     data = &from.:,
                     lib  = finalD
                     );
%mend LossSim_Collect;
