/*-------------------------------------------------------------------
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * %ORVR_Audit_Log(model_uid=)
 *
 * Produces the OpRiskVaR Model Audit Log.
 *
 * Called by %orva_model() - not intended to be called directly!
 *
 * Input:
 *    model_uid - UID for model.
 *
 * Required tables:
 *   vardata.VAR_USER
 *   varprj.VAR_AUDIT
 *
 *-------------------------------------------------------------------*/
%macro ORVR_Audit_Log( model_uid = );
   footnote;
   %let _title_ = %sysfunc(sasmsg(&orvrds, audl_rep_aud_log_title, noquote));
   %orv_clearReportTitles;
   %let _title_5 = &_title_;
   %let proclabel = &_title_5;
   title2 &titleSpec2 "&_title_";
   title3 &titleSpec3 "(&project)";
   title4 &titleSpec3 "(&author)";
   title5 &titleSpec4 "&_title_5";

   *** check for no data;
   %let dsid = %sysfunc(open(varprj.VAR_AUDIT, I));
   %let nobs = %sysfunc(attrn(&dsid, nlobs));
   %let rc   = %sysfunc(close(&dsid));

   %if &nobs = 0 %then %do;
      %let empty_msg = %sysfunc(sasmsg(&orvrds, audl_no_audits_header, noquote));
      ods noproctitle;
      ods proclabel = "&proclabel";
      %orv_emptyReport(&empty_msg.);
      %return;
   %end;

   ***** create VarUser table;
   proc sql noprint;
     create table VarUser as
       select VAR_USER_UID, LAST_NAME, MIDDLE_NAME, FIRST_NAME
       from vardata.VAR_USER;
   quit;
   data VarUser;
      set VarUser;
      length FULL_NAME $ 128;
      FULL_NAME = strip(FIRST_NAME)||" "||strip(MIDDLE_NAME)||" "||strip(LAST_NAME); 
      drop LAST_NAME MIDDLE_NAME FIRST_NAME;
   run;     
  
   ***** merge VAR_AUDIT and VarUser;
   proc sql noprint;
      create table VarAudit as 
       select * 
       from varprj.VAR_AUDIT a, VarUser vu
       having a.MODEL_UID = &model_uid and
              a.VAR_USER_UID = vu.VAR_USER_UID;
   quit;

   ***** SParm Iterations Audit;
   %if &outputtype = PDF %then %do;
      ods PDF startpage = yes;
   %end;

   %orv_clearReportTitles;
   %let _title_5 = %sysfunc(sasmsg(&orvrds, audl_log_aud_sev_fit_title, noquote));
   %let proclabel = &_title_5;

   title2 &titleSpec2 "&_title_";
   title3 &titleSpec3 "(&project)";
   title4 &titleSpec3 "(&author)";
   title5 &titleSpec4 "&_title_5";

   data work; 
      set VarAudit; 
      where upcase(TABLE_NAME) eq 'SPARMI'; 
   run;

   %orv_auditCheckEmpty(data = work);

   %if &nobs ne 0 %then %do;
      data work; 
         set work;
         pos = findc(RECORD,'(');
         ccStr = substr(RECORD, 1, pos - 1);
         CCUID = input(ccStr, best.);
         drop pos ccStr;
      run;

      %orv_auditCCMerge(data=work, dataOut=final);
       
      proc template;
         define table sparm_audit;
            header 
                  ccHeader 
                  h2
                  ;       
            column 
                  (CCVVALUE_0) 
                  CCVVALUE_1 
                  %if &THIRD_DIM %then %do; 
                     CCVALUE_2 
                  %end;
                  OLD_VALUE 
                  NEW_VALUE
                  FULL_NAME 
                  AUDIT_TIME 
                  DESCRIPTION 
                  REASON_FOR_CHANGE
                  ;               
            define ccHeader;
               text "%sysfunc(sasmsg(&orvrds, audl_mod_cell_header, noquote))"; 
               start = CCVVALUE_0  space = 1 expand = '-' end = CCVVALUE_1;
            end;                   

            define CCVVALUE_0; 
               header = "&MAPNAME0" blank_dups = on id; 
            end; 

            define CCVVALUE_1; 
               header="&MAPNAME1";                  
            end;      

            %if &THIRD_DIM %then %do;
               define CCVVALUE_2; 
                  header = "&MAPNAME2";                  
               end;      
            %end;
               
            define h2;
               text "%sysfunc(sasmsg(&orvrds, audl_fit_iter_sel_header, noquote))"; 
               start = OLD_VALUE  space = 1 expand = '-' end = NEW_VALUE;
            end;                   

            define OLD_VALUE; 
               header ="%sysfunc(sasmsg(&orvrds, audl_old_header, noquote))"; 
            end; 

            define NEW_VALUE;                
               header ="%sysfunc(sasmsg(&orvrds, audl_new_header, noquote))"; 
               cellstyle 1 as data {foreground = red font_style = italic font_weight = bold};
            end;

            define FULL_NAME;         
               header = "%sysfunc(sasmsg(&orvrds, audl_user_header, noquote))";              
            end; 

            define AUDIT_TIME;        
               header = "%sysfunc(sasmsg(&orvrds, audl_time_header, noquote))";
               format = &DATETIMEFMT;
            end; 

            define REASON_FOR_CHANGE; 
               header = "%sysfunc(sasmsg(&orvrds, audl_reason_change_header, noquote))"; 
            end; 

            define DESCRIPTION;       
               header = "%sysfunc(sasmsg(&orvrds, audl_audit_event_header, noquote))";       
            end;   
         end;  
      run;

      ods noproctitle;
      ods proclabel = "&proclabel";

      data _NULL_; 
         set final;
         file print ods = (template = 'sparm_audit');
         put _ods_;
      run;  
   %end;

   ***** Primary Insurance Audit;
   %if &outputtype = PDF %then %do;
      ods PDF startpage = yes;
   %end;

   %orv_clearReportTitles;
   %let _title_5 = %sysfunc(sasmsg(&orvrds, audl_log_aud_prim_insur_over_title, noquote));
   %let proclabel = &_title_5;
   title2 &titleSpec2 "&_title_";
   title3 &titleSpec3 "(&project)";
   title4 &titleSpec3 "(&author)";
   title5 &titleSpec4 "&_title_5";

   data work; 
      set VarAudit; 
      where upcase(TABLE_NAME) eq 'INSURE_1'; 
   run;

   %orv_auditCheckEmpty(data = work);

   %if &nobs ne 0 %then %do;
      data work; set work;
         /* get CCUID before ',' */
         pos = findc(RECORD,',');
         tempStr = substr(RECORD,1,pos-1);
         CCUID = input(strip(tempStr),best.);
         /* get InsuranceUID after ',' and before '('*/
         tempStr = substr(RECORD,pos+1);
         pos = findc(tempStr,'(');
         tempStr = substr(tempStr,1,pos-1);
         INSURANCE_UID = input(strip(tempStr),best.);
         if INSURANCE_UID ne 0 then output; 
         drop pos tempStr;
      run;

      %orv_auditCCMerge(data=work, dataOut=work2);

      proc sql;
         create table final as
         select w.*, i.INSURANCE_NAME 
         from work2 as w left join &varorg..INSURANCE as i
         on w.INSURANCE_UID = i.INSURANCE_UID; 
      quit;

      proc sort data = final; 
         by 
            CCUID 
            INSURANCE_UID 
            AUDIT_TIME
            ; 
      run;
       
      proc template;
         define table pinsure_audit;
            header ccHeader h2;       
            column 
                     (CCVVALUE_0) 
                     CCVVALUE_1 
                     %if &THIRD_DIM %then %do; 
                        CCVALUE_2 
                     %end;
                     INSURANCE_NAME 
                     OLD_VALUE 
                     NEW_VALUE
                     FULL_NAME 
                     AUDIT_TIME 
                     DESCRIPTION 
                     REASON_FOR_CHANGE
                     ;               
            define ccHeader;
               text "%sysfunc(sasmsg(&orvrds, audl_mod_cell_header, noquote))"; 
               start = CCVVALUE_0  space = 1 expand = '-' end = CCVVALUE_1;
            end;                   

            define CCVVALUE_0; 
               header = "&MAPNAME0" blank_dups = on id; 
            end; 

            define CCVVALUE_1; 
               header = "&MAPNAME1";                  
            end;      

            %if &THIRD_DIM %then %do;
               define CCVVALUE_2; 
                  header = "&MAPNAME2";                  
               end;      
            %end;

            define h2;
               text "%sysfunc(sasmsg(&orvrds, audl_prob_ins_header, noquote))"; 
               start = OLD_VALUE  space = 1 expand = '-' end = NEW_VALUE;
            end;                   

            define OLD_VALUE; 
               header = "%sysfunc(sasmsg(&orvrds, audl_old_header, noquote))"; 
            end; 

            define NEW_VALUE;                
               header = "%sysfunc(sasmsg(&orvrds, audl_new_header, noquote))"; 
               cellstyle 1 as data {foreground = red font_style = italic font_weight = bold};
            end;

            define FULL_NAME;         
               header = "%sysfunc(sasmsg(&orvrds, audl_user_header, noquote))";              
            end; 

            define AUDIT_TIME;        
               header = "%sysfunc(sasmsg(&orvrds, audl_time_header, noquote))";
               format = &DATETIMEFMT;
            end; 

            define REASON_FOR_CHANGE; 
               header = "%sysfunc(sasmsg(&orvrds, audl_reason_change_header, noquote))"; 
            end; 

            define DESCRIPTION;       
               header = "%sysfunc(sasmsg(&orvrds, audl_audit_event_header, noquote))";       
            end;  

            define INSURANCE_NAME;    
               header = "%sysfunc(sasmsg(&orvrds, audl_policy_header, noquote))"; 
            end;  
         end;  
      run;

      ods noproctitle;
      ods proclabel = "&proclabel";

      data _NULL_; 
         set final;
         file print ods = (template = 'pinsure_audit');
         put _ods_;
      run;  
   %end;

   ***** Secondary Insurance Audit;
   %if &outputtype = PDF %then %do;
      ods PDF startpage = yes;
   %end;

   %orv_clearReportTitles;
   %let _title_5 = = %sysfunc(sasmsg(&orvrds, audl_log_aud_sec_insur_over_title, noquote));
   %let proclabel = &_title_5;

   title2 &titleSpec2 "&_title_";
   title3 &titleSpec3 "(&project)";
   title4 &titleSpec3 "(&author)";
   title5 &titleSpec4 "&_title_5";

   data work; 
      set VarAudit; 
      where upcase(TABLE_NAME) eq 'INSURE_2'; 
   run;

   %orv_auditCheckEmpty(data = work);

   %if &nobs ne 0 %then %do;
      data work; 
         set work;
         pos = findc(RECORD,',');
         tempStr = substr(RECORD,1,pos-1);
         CCUID = input(strip(tempStr),best.);
         /* get InsuranceUID_1 after ',' and before ','*/
         tempStr = substr(RECORD,pos+1);
         pos = findc(tempStr,',');
         tempStr2 = substr(tempStr,1,pos-1);
         INSURANCE_UID_1 = input(strip(tempStr2),best.);
         /* get InsuranceUID_2 after ',' and before '('*/
         tempStr = substr(tempStr,pos+1);
         pos = findc(tempStr,'(');
         tempStr = substr(tempStr,1,pos-1);
         INSURANCE_UID_2 = input(strip(tempStr),best.);
         drop pos tempStr tempStr2;
      run;

      %orv_auditCCMerge(data = work, dataOut = work2);

      proc sql;
         create table work as
         select w.*, i.INSURANCE_NAME 
         from work2 as w left join &varorg..INSURANCE as i
         on w.INSURANCE_UID_1 = i.INSURANCE_UID; 
      quit;

      data work; 
         set work (rename = (INSURANCE_NAME = INSURANCE_NAME_1)); 
      run;

      proc sql;
         create table final as
         select w.*, i.INSURANCE_NAME 
         from work as w left join &varorg..INSURANCE as i
         on w.INSURANCE_UID_2 = i.INSURANCE_UID; 
      quit;
       
      proc sort data = final; 
         by 
            CCUID 
            INSURANCE_UID_1 
            INSURANCE_UID_2 
            AUDIT_TIME
            ; 
      run;

      proc template;
         define table sinsure_audit;
            column 
                  (CCVVALUE_0) 
                  CCVVALUE_1 
                  %if &THIRD_DIM %then %do; 
                     CCVALUE_2 
                  %end;
                  INSURANCE_NAME_1 
                  INSURANCE_NAME 
                  OLD_VALUE 
                  NEW_VALUE
                  FULL_NAME 
                  AUDIT_TIME 
                  DESCRIPTION 
                  REASON_FOR_CHANGE
                  ;               
            header 
                     ccHeader 
                     policyHeader 
                     probHeader
                     ;       
            define ccHeader;
               text "%sysfunc(sasmsg(&orvrds, audl_mod_cell_header, noquote))"; 
               start = CCVVALUE_0  space = 1 expand = '-' end = CCVVALUE_1;
            end;                   

            define CCVVALUE_0; 
               header = "&MAPNAME0" blank_dups = on id; 
            end; 

            define CCVVALUE_1; 
               header = "&MAPNAME1";                  
            end;      

            %if &THIRD_DIM %then %do;
               define CCVVALUE_2; 
                  header = "&MAPNAME2";                  
               end;      
            %end;

            define policyHeader;
               text "%sysfunc(sasmsg(&orvrds, audl_policy_header, noquote))"; 
               start = INSURANCE_NAME_1  space = 1 expand = '-' end = INSURANCE_NAME;
            end;                   

            define INSURANCE_NAME_1; 
               header = "%sysfunc(sasmsg(&orvrds, audl_pri_header, noquote))"; 
            end; 

            define INSURANCE_NAME;   
               header = "%sysfunc(sasmsg(&orvrds, audl_sec_header, noquote))"; 
            end;  

            define probHeader;
               text "%sysfunc(sasmsg(&orvrds, audl_prob_sec_insure_header, noquote))"; 
               start = OLD_VALUE  space = 1 expand = '-' end = NEW_VALUE;
            end;                   

            define OLD_VALUE; 
               header = "%sysfunc(sasmsg(&orvrds, audl_old_header, noquote))"; 
            end; 

            define NEW_VALUE; 
               header = "%sysfunc(sasmsg(&orvrds, audl_new_header, noquote))"; 
               cellstyle 1 as data {foreground = red font_style = italic font_weight = bold};
            end;

            define FULL_NAME;         
               header = "%sysfunc(sasmsg(&orvrds, audl_user_header, noquote))";              
            end; 

            define AUDIT_TIME;        
               header = "%sysfunc(sasmsg(&orvrds, audl_time_header, noquote))";
               format = &DATETIMEFMT;
            end; 

            define REASON_FOR_CHANGE; 
               header = "%sysfunc(sasmsg(&orvrds, audl_reason_change_header, noquote))"; 
            end; 

            define DESCRIPTION;       
               header = "%sysfunc(sasmsg(&orvrds, audl_audit_event_header, noquote))";       
            end;   
         end;  
      run;

      ods noproctitle;
      ods proclabel = "&proclabel";

      data _NULL_; 
         set final;
         file print ods = (template = 'sinsure_audit');
         put _ods_;
      run;  
   %end;
  
   ***** cleanup temporary datasets;
   proc sql noprint;
      drop table VarAudit;
      drop table final;
      drop table work;
   quit;
%mend ORVR_Audit_Log;

%macro orv_auditCheckEmpty(data=);
   %let nobs = 0;
   %let dsid = %sysfunc(open(&data, I));

   %if &dsid ^= 0 %then %do;
       %let nobs = %sysfunc(attrn(&dsid, nlobs));
       %let rc   = %sysfunc(close(&dsid));
   %end;

   %if &nobs = 0 %then %do;
     %orv_emptyReport(%sysfunc(sasmsg(&orvrds, audl_no_audits_header, noquote)));
   %end;  
%mend orv_auditCheckEmpty;

*** get Names for Cross Classification Variables;
%macro orv_auditCCMerge(data =, dataOut =);
   proc sql;
      CREATE TABLE 
                     &dataOut 
      AS SELECT 
                     DISTINCT * 
      FROM 
                     &data f,
                     varprj.cc cc,
                     varprj.ccmapval_0 ccv0,
                     varprj.ccmapval_1 ccv1
                     %if &THIRD_DIM %then %do; 
                        ,varprj.ccmapval_2 ccv2 
                     %end;
      WHERE 
                     f.CCUID = cc.CCUID 
         AND
                     cc.CCVUID_0 = ccv0.CCVUID_0 
         AND 
                     cc.CCVUID_1 = ccv1.CCVUID_1 
                     %if &THIRD_DIM %then %do; 
                        AND cc.CCVUID_2 = ccv2.CCVUID_2 
                     %end;
                     ; 
   quit;
%mend orv_auditCCMerge;
