/* --------------------------------------------------------------
|  Copyright SAS Institute 2014
|  Name: orvu_project_short_name
|
|  Purpose: Test if the project name can be used as a filesystem path
|     name. It can only be used when it has English alphabet characters and 
|     digits and underscores for its characters. 
|
|  Return: The project name with underscores replacing spaces if it 
|     has all valid characters or an empty string if it contains at least
|     on invalid character. The result is returned in the &project_name_short 
|     macro variable.
|
|  Inputs:
|       project_name - The name of the project
|
|  Example:
|    %orvu_project_short_name( "my cool project"); 
|            &project_name_short. = my_cool_project
|
|    %orvu_project_short_name( "pequeo"); 
|            &project_name_short. = 
|
--------------------------------------------------------------------*/

%macro orvu_project_name_get(project_name = );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;   
      %put #orvu_project_short_name(project_name = &project_name.);
   %end;

   data _NULL_;
      length valid_chars $ 37;
	  valid_chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
	 
      length project_name $ 20; /* Purposeful truncation */
      project_name = translate("&project_name", '_', ' ');

      call symputx( "project_name_short", strip(project_name));

	  len = length(strip(project_name));
      do index = 1 to len;
	      currentChar = upcase(char(project_name, index));
          loc = indexc(valid_chars, currentChar);
		  if loc eq 0 then do;
	        call symputx("project_name_short", compress(""));
			leave;
		  end; 
      end;
   run;
%mend orvu_project_name_get;
