/* --------------------------------------------------------------
|  Copyright SAS Institute 2012
|  Name: orvu_unzip
|  Purpose: Unzips a zip file to a directory 
|  Inputs:
|       directory  - The fully qualified directory name
|       zipfile    - The fully qualified zip file name
|  Details:
|     orvu_unzip recursively unzips up all files and folder in the specified
|  zip file. The archive must have been created using the macro %orvu_zip().
|  The program will run on any windows or unix box. Furthermore
|  the package tools move data sets in a portable manner - indexes are
|  preserved. 
|
|
|  Example:
|    %orvu_unzip( c:\temp\out.zip, c:\temp\foo );
--------------------------------------------------------------------*/
%macro orvu_unzip( 
                  zipfile, 
                  directory
                  );
   %if (%upcase(&ORV_DEBUG.) EQ TRUE) %then %do;   
      %put #orvu_unzip(;
      %put .            zipfile   = &zipfile.,; 
      %put .            directory = &directory.;
      %put .            );
   %end;

   proc fcmp ;
      function get_dataset( entryid, dataset $, dir $);
         length dir $256;
         rc = libname("__tmp__", trim(dir));
         CALL RETRIEVE_DATASET(entryid, "__tmp__", trim(scan(dataset,1,'.')), rc);

         rc2 = libname("__tmp__");
         return(rc);
      endsub;

      function get_text( entryid, filename $, dir $ );
         length fn dir $256;
         fn = "filename:"||trim(dir)||"/"||trim(filename);
         CALL RETRIEVE_FILE(entryId, fn, rc);
  
         return(rc);
      endsub;

      function get_package( packageId, num, folder $, dir $ );
         length path msg desc dir $256 type uSpec $32;

         /* create a folder to put this package in */
         path = dcreate( trim(folder), trim(dir) || "/");
         if path = "" then do;
            put "&orvar_error. Can't create " folder "  in directory" dir;
            return(-1);
         end;

         do i=1 to num;
            nv = '';
            rc = 0;
            CALL ENTRY_NEXT(packageId, entryid, type, 
               uSpec, desc, nv,  rc);
           if rc ne 0 then do;
              msg = sysmsg();
              put msg ; 
              put "WARNING: Skipping this package due to a package problem.";   
              rc = 0;             
           end;
           if type = "NESTED_PACKAGE" then do;
              nv = '';
              CALL RETRIEVE_NESTED(entryId, npackageId, num, desc, dt, nv, rc); 
              if rc ne 0 then do;
                  msg = sysmsg();
                  put msg ; 
                  put "WARNING: Skipping this package due to a package problem."  ; 
                  rc = 0;              
              end;
              else
                  rc = get_package( npackageId, num, desc, path );
           end;
           else
           if type = "DATASET" then 
               rc = get_dataset( entryid, desc, path );
           else
           if type = "TEXT" or type = "BINARY" then 
              rc = get_text( entryid, desc, path );
           
           if rc ne 0 then do;
               msg = sysmsg();
               put msg;
             return(rc);
            end;
         end;
         CALL PACKAGE_TERM(packageId, rc);
         return(0);
      endsub;

      subroutine get_root( full $, base $, entry $);
         outargs base, entry;
         length  base entry  $ 256;

         entry = scan( full,-1, '/\');
         ln = length(full);
         if substr(full,ln,1) = '/' or substr(full,ln,1) = '\' then
              ln = ln-1;
         base = substr(full, 1, ln-length(entry)-1);
      endsub;

      length  base entry msg $ 256; 
      /* Main program ---------------------------*/
      file log;
      length type desc $32;
      plist = 0;
      archiveName = "&zipfile";
      rc    = 0;
      total = 0;

      CALL RETRIEVE_PACKAGE(
                            plist, 
                            "FROM_ARCHIVE",
                            archiveName, 
                            total, 
                            rc
                            );
      if rc ne 0 then do;
           msg = sysmsg();
           put msg ;
           put "ERROR: Unable to open archive.";            
      end;
      packageId = 0;
      desc = '';
      num  = 0;
      dt   = 0;
      nv   = '';
      ch   = '';

      CALL PACKAGE_FIRST(
                         plist, 
                         packageId, 
                         num, 
                         desc, 
                         dt, 
                         nv, 
                         ch, 
                         rc
                         );
      if rc ne 0 then do;
           msg = sysmsg();
           put msg ;
           put "ERROR: Unable to process archive.";            
      end;

      call get_root("&directory", base, entry);
      rc = get_package( packageId, num, entry, base);
      CALL PACKAGE_TERM(plist, rc);
   run;
%mend orvu_unzip;
